from hub.util.exceptions import LockedException
import numpy as np
import pytest
import hub
import uuid
import time
import warnings


def test_dataset_locking(s3_ds_generator):
    ds = s3_ds_generator()
    ds.create_tensor("x")
    arr = np.random.random((32, 32))
    ds.x.append(arr)

    # Emulate a different machine
    getnode = uuid.getnode
    uuid.getnode = lambda: getnode() + 1
    hub.core.lock._LOCKS.clear()

    try:

        # Make sure read only warning is raised
        with pytest.warns(UserWarning):
            ds = s3_ds_generator()
            np.testing.assert_array_equal(arr, ds.x[0].numpy())
        assert ds.read_only == True

        # No warnings if user requests read only mode
        with warnings.catch_warnings(record=True) as ws:
            ds = s3_ds_generator(read_only=True)
            np.testing.assert_array_equal(arr, ds.x[0].numpy())
        assert not ws

        DATASET_LOCK_VALIDITY = hub.constants.DATASET_LOCK_VALIDITY
        # Temporarily set validity to 1 second so we dont have to wait too long.
        hub.constants.DATASET_LOCK_VALIDITY = 1
        # Wait for lock to expire.
        time.sleep(1.1)

        try:
            ds = s3_ds_generator()
            np.testing.assert_array_equal(arr, ds.x[0].numpy())
            assert ds.read_only == False
        finally:
            hub.constants.DATASET_LOCK_VALIDITY = DATASET_LOCK_VALIDITY
    finally:
        uuid.getnode = getnode



import numpy as np
from hub.core.meta.encode.byte_positions import BytePositionsEncoder
from .common import assert_encoded


def _validate_bp(enc: BytePositionsEncoder):
    """Helps validate that tests have proper initial encoder states."""

    expected = np.array(enc._encoded)
    enc._post_process_state(0)
    actual = enc._encoded

    np.testing.assert_array_equal(expected, actual)


def test_update_no_change():
    enc = BytePositionsEncoder([[8, 0, 10], [4, 88, 15]])
    _validate_bp(enc)

    enc[0] = 8
    assert_encoded(enc, [[8, 0, 10], [4, 88, 15]])

    enc[1] = 8
    assert_encoded(enc, [[8, 0, 10], [4, 88, 15]])

    enc[10] = 8
    assert_encoded(enc, [[8, 0, 10], [4, 88, 15]])

    enc[11] = 4
    assert_encoded(enc, [[8, 0, 10], [4, 88, 15]])

    enc[12] = 4
    assert_encoded(enc, [[8, 0, 10], [4, 88, 15]])

    enc[15] = 4
    assert_encoded(enc, [[8, 0, 10], [4, 88, 15]])

    assert enc.num_samples == 16


def test_update_squeeze_trivial():
    enc = BytePositionsEncoder([[4, 0, 9], [2, 40, 10], [4, 42, 29]])
    _validate_bp(enc)

    enc[10] = 4
    assert_encoded(enc, [[4, 0, 29]])

    assert enc.num_samples == 30


def test_update_squeeze_complex():
    enc = BytePositionsEncoder(
        [[2, 0, 5], [4, 12, 9], [2, 28, 10], [4, 30, 29], [2, 106, 100]]
    )
    _validate_bp(enc)

    enc[10] = 4
    assert_encoded(enc, [[2, 0, 5], [4, 12, 29], [2, 108, 100]])

    assert enc.num_samples == 101



import warnings
import numpy as np
from pathlib import Path
import os
import glob
import PIL  # type: ignore
from typing import Dict, List, Sequence, Tuple, Union

from hub.util.auto import ingestion_summary
from hub.util.exceptions import (
    InvalidPathException,
    TensorInvalidSampleShapeError,
)
from hub.core.dataset import Dataset

from tqdm import tqdm  # type: ignore

from .base import UnstructuredDataset

import hub

IMAGES_TENSOR_NAME = "images"
LABELS_TENSOR_NAME = "labels"


def _get_file_paths(directory: Path, relative_to: Union[str, Path] = "") -> List[Path]:
    g = glob.glob(os.path.join(directory, "**"), recursive=True)
    file_paths = []
    for path_str in g:
        if os.path.isfile(path_str):
            path = Path(path_str)
            if relative_to:
                relative_path = Path(path).relative_to(directory)
            else:
                relative_path = path
            file_paths.append(relative_path)
    return file_paths


def _class_name_from_path(path: Path) -> str:
    return path.parts[-2]


def _set_name_from_path(path: Path) -> str:
    return path.parts[-3]


class ImageClassification(UnstructuredDataset):
    def __init__(self, source: str):
        """Convert an unstructured dataset to a structured dataset.

        Note:
            Currently only supports computer vision (image) datasets.

        Args:
            source (str): The full path to the dataset.
                Can be a Hub cloud path of the form hub://username/datasetname. To write to Hub cloud datasets, ensure that you are logged in to Hub (use 'activeloop login' from command line)
                Can be a s3 path of the form s3://bucketname/path/to/dataset. Credentials are required in either the environment or passed to the creds argument.
                Can be a local file system path of the form ./path/to/dataset or ~/path/to/dataset or path/to/dataset.
                Can be a memory path of the form mem://path/to/dataset which doesn't save the dataset but keeps it in memory instead. Should be used only for testing as it does not persist.

        Raises:
            InvalidPathException: If source is invalid.
        """

        super().__init__(source)

        self._abs_file_paths = _get_file_paths(self.source)
        self._rel_file_paths = _get_file_paths(self.source, relative_to=self.source)
        if len(self._abs_file_paths) == 0:
            raise InvalidPathException(
                f"No files found in {self.source}. Please ensure that the source path is correct."
            )

        self.set_names = self.get_set_names()
        self.class_names = self.get_class_names()

    # TODO: make lazy/memoized property
    def get_set_names(self) -> Tuple[str, ...]:
        # TODO: move outside class
        set_names = set()
        for file_path in self._abs_file_paths:
            set_names.add(_set_name_from_path(file_path))
        return tuple(sorted(set_names))  # TODO: lexicographical sorting

    # TODO: make lazy/memoized property
    def get_class_names(self) -> Tuple[str, ...]:
        # TODO: move outside class
        class_names = set()
        for file_path in self._abs_file_paths:
            class_names.add(_class_name_from_path(file_path))
        return tuple(sorted(class_names))  # TODO: lexicographical sorting

    def structure(  # type: ignore
        self,
        ds: Dataset,
        use_progress_bar: bool = True,
        generate_summary: bool = True,
        image_tensor_args: dict = {},
    ) -> Dataset:
        """Create a structured dataset.

        Args:
            ds (Dataset) : A Hub dataset object.
            use_progress_bar (bool): Defines if the method uses a progress bar. Defaults to True.
            generate_summary (bool): Defines if the method generates ingestion summary. Defaults to True.
            image_tensor_args (dict): Defines the sample compression of the dataset (jpeg or png).

        Returns:
            A hub dataset.

        """

        images_tensor_map = {}
        labels_tensor_map = {}

        use_set_prefix = len(self.set_names) > 1

        for set_name in self.set_names:
            if not use_set_prefix:
                set_name = ""

            images_tensor_name = os.path.join(set_name, IMAGES_TENSOR_NAME)
            labels_tensor_name = os.path.join(set_name, LABELS_TENSOR_NAME)
            images_tensor_map[set_name] = images_tensor_name.replace("\\", "/")
            labels_tensor_map[set_name] = labels_tensor_name.replace("\\", "/")

            # TODO: infer sample_compression
            ds.create_tensor(
                images_tensor_name.replace("\\", "/"),
                htype="image",
                **image_tensor_args,
            )
            ds.create_tensor(
                labels_tensor_name.replace("\\", "/"),
                htype="class_label",
                class_names=self.class_names,
            )

            paths = self._abs_file_paths
            skipped_files: list = []

            iterator = tqdm(
                paths,
                desc='Ingesting "%s" (%i files skipped)'
                % (self.source.name, len(skipped_files)),
                total=len(paths),
                disable=not use_progress_bar,
            )

        with ds, iterator:
            for file_path in iterator:
                try:
                    image = hub.read(file_path)
                except PIL.Image.UnidentifiedImageError:
                    skipped_files.append(file_path.name)
                    iterator.set_description(
                        'Ingesting "%s" (%i files skipped)'
                        % (self.source.name, len(skipped_files))
                    )
                    continue

                class_name = _class_name_from_path(file_path)

                label = np.uint32(self.class_names.index(class_name))

                set_name = _set_name_from_path(file_path) if use_set_prefix else ""

                # TODO: try to get all len(shape)s to match.
                # if appending fails because of a shape mismatch, expand dims (might also fail)
                try:
                    ds[images_tensor_map[set_name]].append(image)

                except TensorInvalidSampleShapeError:
                    im = image.array
                    reshaped_image = np.expand_dims(im, -1)
                    ds[images_tensor_map[set_name]].append(reshaped_image)

                except Exception:
                    skipped_files.append(file_path.name)
                    iterator.set_description(
                        'Ingesting "%s" (%i files skipped)'
                        % (self.source.name, len(skipped_files))
                    )
                    continue

                ds[labels_tensor_map[set_name]].append(label)

            if generate_summary:
                ingestion_summary(str(self.source), skipped_files)
            return ds



import textwrap

from hub.client.config import HUB_REST_ENDPOINT
import click
from humbug.report import Report

from hub.client.client import HubBackendClient
from hub.client.utils import remove_username_from_config, write_token, remove_token
from hub.util.bugout_reporter import (
    save_reporting_config,
    get_reporting_config,
    hub_reporter,
)
from hub.util.exceptions import AuthenticationException


@click.command()
@click.option("--username", "-u", default=None, help="Your Activeloop Username")
@click.option("--password", "-p", default=None, help="Your Activeloop Password")
def login(username: str, password: str):
    """Log in to Activeloop"""
    if not username:
        click.echo("Login to Activeloop using your credentials.")
        click.echo(
            "If you don't have an account, register by using the 'activeloop register' command or by going to "
            f"{HUB_REST_ENDPOINT}/register."
        )
        username = click.prompt("Username")
        password = click.prompt("Password", hide_input=True)
    if not password:
        password = click.prompt(
            f"Please enter password for user {username}", hide_input=True
        )
    username = username.strip()
    password = password.strip()
    try:
        client = HubBackendClient()
        token = client.request_auth_token(username, password)
        write_token(token)
        click.echo("Successfully logged in to Activeloop.")
        reporting_config = get_reporting_config()
        if reporting_config.get("username") != username:
            save_reporting_config(True, username=username)
    except AuthenticationException:
        raise SystemExit("Login failed. Check username and password.")
    except Exception as e:
        raise SystemExit(f"Unable to login: {e}")


@click.command()
def logout():
    """Log out of Activeloop"""
    remove_token()
    remove_username_from_config()
    click.echo("Logged out of Activeloop.")


# TODO: Add how to enable/disable reporting to docs
@click.command()
@click.option("--on/--off", help="Turn crash report on/off")
def reporting(on):
    """Enable or disable sending crash report to Activeloop AI"""
    report = Report(
        title="Consent change",
        tags=hub_reporter.system_tags(),
        content=f"Consent? `{on}`",
    )
    hub_reporter.publish(report)
    save_reporting_config(on)


@click.command()
@click.option("--username", "-u", default=None, help="Your Activeloop Username")
@click.option("--email", "-e", default=None, help="Your Email")
@click.option("--password", "-p", default=None, help="Your Activeloop Password")
def register(username: str, email: str, password: str):
    """Create a new Activeloop user account"""
    click.echo("Thank you for registering for an Activeloop account!")
    if not username:
        click.echo(
            "Enter your details. Your password must be atleast 6 characters long."
        )
        username = click.prompt("Username")
        email = click.prompt("Email")
        password = click.prompt("Password", hide_input=True)
    if not password:
        password = click.prompt(
            f"Please enter the password you would like to associate with {username}",
            hide_input=True,
        )
    if not email:
        email = click.prompt(
            f"Please enter the Email Address you would like to associate with {username}"
        )
    username = username.strip()
    email = email.strip()
    password = password.strip()
    try:
        client = HubBackendClient()
        client.send_register_request(username, email, password)
        token = client.request_auth_token(username, password)
        write_token(token)
        click.echo(
            f"Successfully registered and logged in to Activeloop as {username}."
        )
        consent_message = textwrap.dedent(
            """
            Privacy policy:
            We collect basic system information and crash reports so that we can keep
            improving your experience using Hub to work with your data.
            You can find out more by reading our privacy policy:
                https://www.activeloop.ai/privacy/
            If you would like to opt out of reporting crashes and system information,
            run the following command:
                $ activeloop reporting --off
            """
        )
        click.echo(consent_message)
        save_reporting_config(True, username=username)
    except Exception as e:
        raise SystemExit(f"Unable to register new user: {e}")



import pytest
import hub

enabled_datasets = pytest.mark.parametrize(
    "ds",
    ["memory_ds", "local_ds", "s3_ds", "gcs_ds"],
    indirect=True,
)

enabled_non_gcs_datasets = pytest.mark.parametrize(
    "ds",
    ["memory_ds", "local_ds", "s3_ds"],
    indirect=True,
)

enabled_persistent_dataset_generators = pytest.mark.parametrize(
    "ds_generator",
    ["local_ds_generator", "s3_ds_generator", "gcs_ds_generator"],
    indirect=True,
)


@pytest.fixture
def memory_ds(memory_path):
    return hub.dataset(memory_path)


@pytest.fixture
def local_ds(local_ds_generator):
    return local_ds_generator()


@pytest.fixture
def local_ds_generator(local_path):
    def generate_local_ds(**kwargs):
        return hub.dataset(local_path, **kwargs)

    return generate_local_ds


@pytest.fixture
def s3_ds(s3_ds_generator):
    return s3_ds_generator()


@pytest.fixture
def s3_ds_generator(s3_path):
    def generate_s3_ds(**kwargs):
        return hub.dataset(s3_path, **kwargs)

    return generate_s3_ds


@pytest.fixture
def gcs_ds(gcs_ds_generator):
    return gcs_ds_generator()


@pytest.fixture
def gcs_ds_generator(gcs_path, gcs_creds):
    def generate_gcs_ds():
        return hub.dataset(gcs_path, creds=gcs_creds)

    return generate_gcs_ds


@pytest.fixture
def hub_cloud_ds(hub_cloud_ds_generator):
    return hub_cloud_ds_generator()


@pytest.fixture
def hub_cloud_ds_generator(hub_cloud_path, hub_cloud_dev_token):
    def generate_hub_cloud_ds(**kwargs):
        return hub.dataset(hub_cloud_path, token=hub_cloud_dev_token, **kwargs)

    return generate_hub_cloud_ds


@pytest.fixture
def ds(request):
    """Used with parametrize to use all enabled dataset fixtures."""
    return request.getfixturevalue(request.param)


@pytest.fixture
def ds_generator(request):
    """Used with parametrize to use all enabled persistent dataset generator fixtures."""
    return request.getfixturevalue(request.param)



import numpy as np
from hub.util import shuffle


def test_shuffle(memory_ds):
    ds = memory_ds
    ds.create_tensor("ints", dtype="int64")
    ds.ints.extend(np.arange(10, dtype="int64").reshape((10, 1)))

    np.random.seed(0)
    ds = shuffle(ds)
    expected = [[2], [8], [4], [9], [1], [6], [7], [3], [0], [5]]
    assert ds.ints.numpy().tolist() == expected

    assert ds.ints[0].numpy() == 2



"""
Supported compressions (formats):
    Image : bmp, dib, gif, ico, jpeg, jp2, pcx, png, ppm, sgi, tga, tiff, webp, wmf, xbm
    Audio : flac, mp3, wav
    Video : mp4, mkv, avi
    Bytes : lz4

"""
from PIL import Image  # type: ignore


BYTE_COMPRESSIONS = [
    "lz4",
]


IMAGE_COMPRESSIONS = [
    "bmp",
    "dib",
    "gif",
    "ico",
    "jpeg",
    "jpeg2000",
    "pcx",
    "png",
    "ppm",
    "sgi",
    "tga",
    "tiff",
    "webp",
    "wmf",
    "xbm",
]

VIDEO_COMPRESSIONS = ["mp4", "mkv", "avi"]

AUDIO_COMPRESSIONS = ["mp3", "flac", "wav"]


# Just constants
BYTE_COMPRESSION = "byte"
IMAGE_COMPRESSION = "image"
VIDEO_COMPRESSION = "video"
AUDIO_COMPRESSION = "audio"


COMPRESSION_TYPES = [BYTE_COMPRESSION, IMAGE_COMPRESSION, AUDIO_COMPRESSION]


COMPRESSION_TYPES = [
    BYTE_COMPRESSION,
    IMAGE_COMPRESSION,
    AUDIO_COMPRESSION,
    VIDEO_COMPRESSION,
]

# Pillow plugins for some formats might not be installed:
Image.init()
IMAGE_COMPRESSIONS = [
    c for c in IMAGE_COMPRESSIONS if c.upper() in Image.SAVE and c.upper() in Image.OPEN
]

IMAGE_COMPRESSIONS.insert(0, "apng")

SUPPORTED_COMPRESSIONS = [
    *BYTE_COMPRESSIONS,
    *IMAGE_COMPRESSIONS,
    *AUDIO_COMPRESSIONS,
    *VIDEO_COMPRESSIONS,
]
SUPPORTED_COMPRESSIONS = list(sorted(set(SUPPORTED_COMPRESSIONS)))  # type: ignore
SUPPORTED_COMPRESSIONS.append(None)  # type: ignore

COMPRESSION_ALIASES = {"jpg": "jpeg"}

# If `True`  compression format has to be the same between samples in the same tensor.
# If `False` compression format can   be different between samples in the same tensor.
USE_UNIFORM_COMPRESSION_PER_SAMPLE = True


_compression_types = {}
for c in IMAGE_COMPRESSIONS:
    _compression_types[c] = IMAGE_COMPRESSION
for c in BYTE_COMPRESSIONS:
    _compression_types[c] = BYTE_COMPRESSION
for c in VIDEO_COMPRESSIONS:
    _compression_types[c] = VIDEO_COMPRESSION
for c in AUDIO_COMPRESSIONS:
    _compression_types[c] = AUDIO_COMPRESSION


def get_compression_type(c):
    if c is None:
        return None
    ret = _compression_types.get(c)
    if ret is None and c.upper() in Image.OPEN:
        ret = IMAGE_COMPRESSION
    if ret is None:
        raise KeyError(c)
    return ret



from hub.core.meta.tensor_meta import TensorMeta
from hub.util.exceptions import TensorInvalidSampleShapeError
from hub.util.casting import intelligent_cast
from hub.util.json import HubJsonEncoder, validate_json_object
from hub.core.sample import Sample, SampleValue  # type: ignore
from hub.core.compression import compress_array, compress_bytes
from hub.client import config
from hub.compression import IMAGE_COMPRESSIONS
from typing import List, Optional, Sequence, Union, Tuple, Iterable
import hub
import numpy as np
import struct
import warnings
import json


def infer_chunk_num_bytes(
    version: str,
    shape_info: np.ndarray,
    byte_positions: np.ndarray,
    data: Optional[Union[Sequence[bytes], Sequence[memoryview]]] = None,
    len_data: Optional[int] = None,
) -> int:
    """Calculates the number of bytes in a chunk without serializing it. Used by `LRUCache` to determine if a chunk can be cached.

    Args:
        version: (str) Version of hub library
        shape_info: (numpy.ndarray) Encoded shapes info from the chunk's `ShapeEncoder` instance.
        byte_positions: (numpy.ndarray) Encoded byte positions from the chunk's `BytePositionsEncoder` instance.
        data: (list) `_data` field of the chunk
        len_data: (int, optional) Number of bytes in the chunk

    Returns:
        Length of the chunk when serialized as int
    """
    # NOTE: Assumption: version string contains ascii characters only (ord(c) < 128)
    # NOTE: Assumption: len(version) < 256
    if len_data is None:
        len_data = sum(map(len, data))  # type: ignore
    return len(version) + shape_info.nbytes + byte_positions.nbytes + len_data + 13  # type: ignore


def serialize_chunk(
    version: str,
    shape_info: np.ndarray,
    byte_positions: np.ndarray,
    data: Union[Sequence[bytes], Sequence[memoryview]],
    len_data: Optional[int] = None,
) -> memoryview:
    """Serializes a chunk's headers and data into a single byte stream. This is how the chunk will be written to the storage provider.

    Args:
        version: (str) Version of hub library.
        shape_info: (numpy.ndarray) Encoded shapes info from the chunk's `ShapeEncoder` instance.
        byte_positions: (numpy.ndarray) Encoded byte positions from the chunk's `BytePositionsEncoder` instance.
        data: (list) `_data` field of the chunk.
        len_data: (int, optional) Number of bytes in the chunk.

    Returns:
        Serialized chunk as memoryview.
    """
    nbytes = infer_chunk_num_bytes(version, shape_info, byte_positions, data, len_data)
    flatbuff = bytearray(nbytes)
    offset = write_version(version, flatbuff)
    offset = write_shape_info(shape_info, flatbuff, offset)
    offset = write_byte_positions(byte_positions, flatbuff, offset)
    offset = write_actual_data(data, flatbuff, offset)
    return memoryview(flatbuff)


def write_version(version, buffer) -> int:
    """Writes version info to the buffer, returns offset."""
    len_version = len(version)
    buffer[0] = len_version
    buffer[1 : 1 + len_version] = version.encode("ascii")
    offset = 1 + len_version
    return offset


def write_shape_info(shape_info, buffer, offset) -> int:
    """Writes shape info to the buffer, takes offset into account and returns updated offset."""
    if shape_info.ndim == 1:
        offset += 8
    else:
        buffer[offset : offset + 8] = struct.pack("<ii", *shape_info.shape)
        offset += 8

        buffer[offset : offset + shape_info.nbytes] = shape_info.tobytes()
        offset += shape_info.nbytes
    return offset


def write_byte_positions(byte_positions, buffer, offset) -> int:
    """Writes byte positions info to the buffer, takes offset into account and returns updated offset."""
    if byte_positions.ndim == 1:
        offset += 4
    else:
        buffer[offset : offset + 4] = byte_positions.shape[0].to_bytes(4, "little")
        offset += 4

        buffer[offset : offset + byte_positions.nbytes] = byte_positions.tobytes()
        offset += byte_positions.nbytes
    return offset


def write_actual_data(data, buffer, offset) -> int:
    """Writes actual chunk data to the buffer, takes offset into account and returns updated offset"""
    for byts in data:
        n = len(byts)
        buffer[offset : offset + n] = byts
        offset += n
    return offset


def deserialize_chunk(
    byts: Union[bytes, memoryview], copy: bool = True
) -> Tuple[str, np.ndarray, np.ndarray, memoryview]:
    """Deserializes a chunk from the serialized byte stream. This is how the chunk can be accessed/modified after it is read from storage.

    Args:
        byts: (bytes) Serialized chunk.
        copy: (bool) If true, this function copies the byts while deserializing incase byts was a memoryview.

    Returns:
        Tuple of:
        hub version used to create the chunk,
        encoded shapes info as numpy array,
        encoded byte positions as numpy array,
        chunk data as memoryview.
    """
    incoming_mview = isinstance(byts, memoryview)
    byts = memoryview(byts)

    enc_dtype = np.dtype(hub.constants.ENCODING_DTYPE)
    itemsize = enc_dtype.itemsize

    # Read version
    len_version = byts[0]
    version = str(byts[1 : 1 + len_version], "ascii")
    offset = 1 + len_version

    # Read shape info
    shape_info_nrows, shape_info_ncols = struct.unpack("<ii", byts[offset : offset + 8])
    offset += 8
    shape_info_nbytes = shape_info_nrows * shape_info_ncols * itemsize
    if shape_info_nbytes == 0:
        shape_info = np.array([], dtype=enc_dtype)
    else:
        shape_info = (
            np.frombuffer(byts[offset : offset + shape_info_nbytes], dtype=enc_dtype)
            .reshape(shape_info_nrows, shape_info_ncols)
            .copy()
        )
        offset += shape_info_nbytes

    # Read byte positions
    byte_positions_rows = int.from_bytes(byts[offset : offset + 4], "little")
    offset += 4
    byte_positions_nbytes = byte_positions_rows * 3 * itemsize
    if byte_positions_nbytes == 0:
        byte_positions = np.array([], dtype=enc_dtype)
    else:
        byte_positions = (
            np.frombuffer(
                byts[offset : offset + byte_positions_nbytes], dtype=enc_dtype
            )
            .reshape(byte_positions_rows, 3)
            .copy()
        )
        offset += byte_positions_nbytes

    # Read data
    data = byts[offset:]
    if incoming_mview and copy:
        data = memoryview(bytes(data))
    return version, shape_info, byte_positions, data  # type: ignore


def serialize_chunkids(version: str, ids: Sequence[np.ndarray]) -> memoryview:
    """Serializes chunk ID encoders into a single byte stream. This is how the encoders will be written to the storage provider.

    Args:
        version: (str) Version of hub library.
        ids: (list) Encoded chunk ids from a `ChunkIdEncoder` instance.

    Returns:
        Serialized chunk ids as memoryview.
    """
    len_version = len(version)
    flatbuff = bytearray(1 + len_version + sum([x.nbytes for x in ids]))

    # Write version
    len_version = len(version)
    flatbuff[0] = len_version
    flatbuff[1 : 1 + len_version] = version.encode("ascii")
    offset = 1 + len_version

    # Write ids
    for arr in ids:
        flatbuff[offset : offset + arr.nbytes] = arr.tobytes()
        offset += arr.nbytes

    return memoryview(flatbuff)


def deserialize_chunkids(byts: Union[bytes, memoryview]) -> Tuple[str, np.ndarray]:
    """Deserializes a chunk ID encoder from the serialized byte stream. This is how the encoder can be accessed/modified after it is read from storage.

    Args:
        byts: (bytes) Serialized chunk ids.

    Returns:
        Tuple of:
        hub version used to create the chunk,
        encoded chunk ids as memoryview.
    """
    byts = memoryview(byts)
    enc_dtype = np.dtype(hub.constants.ENCODING_DTYPE)

    # Read version
    len_version = byts[0]
    version = str(byts[1 : 1 + len_version], "ascii")
    offset = 1 + len_version

    # Read chunk ids
    ids = np.frombuffer(byts[offset:], dtype=enc_dtype).reshape(-1, 2).copy()
    return version, ids


def _serialize_input_sample(
    sample: SampleValue,
    sample_compression: Optional[str],
    expected_dtype: str,
    htype: str,
) -> Tuple[bytes, Tuple[int]]:
    """Converts the incoming sample into a buffer with the proper dtype and compression."""

    if htype in ("json", "list"):
        if isinstance(sample, np.ndarray):
            if htype == "list":
                if sample.dtype == object:
                    sample = list(sample)
                else:
                    sample = sample.tolist()
            elif htype == "json":
                if sample.ndim == 0:
                    sample = sample.tolist()  # actually returns dict
                elif sample.dtype == object:
                    sample = list(sample)
                else:
                    sample = sample.tolist()
        validate_json_object(sample, expected_dtype)
        byts = json.dumps(sample, cls=HubJsonEncoder).encode()
        if sample_compression:
            byts = compress_bytes(byts, compression=sample_compression)
        shape = (len(sample),) if htype == "list" else (1,)
        return byts, shape
    elif htype == "text":
        if isinstance(sample, np.ndarray):
            sample = sample.tolist()
        if not isinstance(sample, str):
            raise TypeError("Expected str, received: " + str(sample))
        byts = sample.encode()
        if sample_compression:
            byts = compress_bytes(byts, compression=sample_compression)
        return byts, (1,)

    if isinstance(sample, Sample):
        if (
            sample_compression
            and hub.compression.get_compression_type(sample_compression) == "byte"
        ):
            # Byte compressions don't store dtype info, so have to cast incoming samples to expected dtype
            arr = intelligent_cast(sample.array, expected_dtype, htype)
            sample = Sample(array=arr)
        buffer = sample.compressed_bytes(sample_compression)
        shape = sample.shape
    else:
        sample = intelligent_cast(sample, expected_dtype, htype)
        shape = sample.shape

        if sample_compression is not None:
            buffer = compress_array(sample, sample_compression)
        else:
            buffer = sample.tobytes()

    if len(shape) == 0:
        shape = (1,)

    return buffer, shape


def _check_input_samples_are_valid(
    num_bytes: List[int],
    shapes: List[Tuple[int]],
    min_chunk_size: int,
    sample_compression: Optional[str],
):
    """Iterates through all buffers/shapes and raises appropriate errors."""

    expected_dimensionality = None
    for nbytes, shape in zip(num_bytes, shapes):
        # check that all samples have the same dimensionality
        if expected_dimensionality is None:
            expected_dimensionality = len(shape)

        if nbytes > min_chunk_size:
            msg = f"Sorry, samples that exceed minimum chunk size ({min_chunk_size} bytes) are not supported yet (coming soon!). Got: {nbytes} bytes."
            if sample_compression is None:
                msg += "\nYour data is uncompressed, so setting `sample_compression` in `Dataset.create_tensor` could help here!"
            raise NotImplementedError(msg)

        if len(shape) != expected_dimensionality:
            raise TensorInvalidSampleShapeError(shape, expected_dimensionality)


def serialize_input_samples(
    samples: Union[Sequence[SampleValue], np.ndarray],
    meta: TensorMeta,
    min_chunk_size: int,
) -> Tuple[Union[memoryview, bytearray], List[int], List[Tuple[int]]]:
    """Casts, compresses, and serializes the incoming samples into a list of buffers and shapes.

    Args:
        samples (Union[Sequence[SampleValue], np.ndarray]): Ssequence of samples.
        meta (TensorMeta): Tensor meta. Will not be modified.
        min_chunk_size (int): Used to validate that all samples are appropriately sized.

    Raises:
        ValueError: Tensor meta should have it's dtype set.
        NotImplementedError: When extending tensors with Sample insatances.
        TypeError: When sample type is not understood.

    Returns:
        List[Tuple[memoryview, Tuple[int]]]: Buffers and their corresponding shapes for the input samples.
    """

    if meta.dtype is None:
        raise ValueError("Dtype must be set before input samples can be serialized.")

    sample_compression = meta.sample_compression
    chunk_compression = meta.chunk_compression
    dtype = meta.dtype
    htype = meta.htype

    if sample_compression or not hasattr(samples, "dtype"):
        buff = bytearray()
        nbytes = []
        shapes = []
        expected_dim = len(meta.max_shape)
        is_convert_candidate = (
            (htype == "image")
            or sample_compression in IMAGE_COMPRESSIONS
            or chunk_compression in IMAGE_COMPRESSIONS
        )

        for sample in samples:
            byts, shape = _serialize_input_sample(
                sample, sample_compression, dtype, htype
            )
            if (
                isinstance(sample, Sample)
                and is_convert_candidate
                and hub.constants.CONVERT_GRAYSCALE
            ):
                if not expected_dim:
                    expected_dim = len(shape)
                if len(shape) == 2 and expected_dim == 3:
                    warnings.warn(
                        f"Grayscale images will be reshaped from (H, W) to (H, W, 1) to match tensor dimensions. This warning will be shown only once."
                    )
                    shape += (1,)  # type: ignore[assignment]
            buff += byts
            nbytes.append(len(byts))
            shapes.append(shape)
    elif (
        isinstance(samples, np.ndarray)
        or np.isscalar(samples)
        or isinstance(samples, Sequence)
    ):
        samples = intelligent_cast(samples, dtype, htype)
        buff = memoryview(samples.tobytes())  # type: ignore
        if len(samples):
            shape = samples[0].shape
            nb = samples[0].nbytes
            if not shape:
                shape = (1,)
        else:
            shape = ()  # type: ignore
            nb = 0
        nbytes = [nb] * len(samples)
        shapes = [shape] * len(samples)
    elif isinstance(samples, Sample):
        # TODO
        raise NotImplementedError(
            "Extending with `Sample` instance is not supported yet."
        )
    else:
        raise TypeError(f"Cannot serialize samples of type {type(samples)}")
    _check_input_samples_are_valid(nbytes, shapes, min_chunk_size, sample_compression)
    return buff, nbytes, shapes



import hub
from hub.core.fast_forwarding import ffw_tensor_meta
from typing import Any, Callable, Dict, List, Tuple
import numpy as np
from hub.util.exceptions import (
    TensorMetaInvalidHtype,
    TensorMetaInvalidHtypeOverwriteValue,
    TensorMetaInvalidHtypeOverwriteKey,
    TensorMetaMissingRequiredValue,
    TensorMetaMutuallyExclusiveKeysError,
    UnsupportedCompressionError,
    TensorInvalidSampleShapeError,
)
from hub.util.json import validate_json_schema
from hub.constants import (
    REQUIRE_USER_SPECIFICATION,
    UNSPECIFIED,
)
from hub.compression import (
    COMPRESSION_ALIASES,
    get_compression_type,
    AUDIO_COMPRESSION,
    BYTE_COMPRESSION,
    IMAGE_COMPRESSION,
    VIDEO_COMPRESSION,
)
from hub.htype import (
    HTYPE_CONFIGURATIONS,
)
from hub.htype import HTYPE_CONFIGURATIONS, REQUIRE_USER_SPECIFICATION, UNSPECIFIED
from hub.core.meta.meta import Meta


class TensorMeta(Meta):
    htype: str
    dtype: str
    min_shape: List[int]
    max_shape: List[int]
    length: int
    sample_compression: str
    chunk_compression: str
    max_chunk_size: int

    def __init__(
        self,
        htype: str = UNSPECIFIED,
        **kwargs,
    ):
        """Tensor metadata is responsible for keeping track of global sample metadata within a tensor.

        Note:
            Tensor metadata that is automatically synchronized with `storage`. For more details, see the `Meta` class.
            Auto-populates `required_meta` that `Meta` accepts as an argument.

        Args:
            htype (str): All tensors require an `htype`. This determines the default meta keys/values.
            **kwargs: Any key that the provided `htype` has can be overridden via **kwargs. For more information, check out `hub.htype`.
        """

        if htype != UNSPECIFIED:
            _validate_htype_exists(htype)
            _validate_htype_overwrites(htype, kwargs)
            _replace_unspecified_values(htype, kwargs)
            _validate_required_htype_overwrites(htype, kwargs)
            _format_values(htype, kwargs)

            required_meta = _required_meta_from_htype(htype)
            required_meta.update(kwargs)

            self._required_meta_keys = tuple(required_meta.keys())
            self.__dict__.update(required_meta)
        else:
            self._required_meta_keys = tuple()

        super().__init__()

    def set_dtype(self, dtype: np.dtype):
        """Should only be called once."""
        ffw_tensor_meta(self)

        if self.dtype is not None:
            raise ValueError(
                f"Tensor meta already has a dtype ({self.dtype}). Incoming: {dtype.name}."
            )

        if self.length > 0:
            raise ValueError("Dtype was None, but length was > 0.")

        self.dtype = dtype.name

    def update_shape_interval(self, shape: Tuple[int, ...]):
        ffw_tensor_meta(self)

        if not self.min_shape:  # both min_shape and max_shape are set together
            self.min_shape = list(shape)
            self.max_shape = list(shape)
        else:
            expected_dims = len(self.min_shape)

            if len(shape) != expected_dims:
                raise TensorInvalidSampleShapeError(shape, len(self.min_shape))

            for i, dim in enumerate(shape):
                self.min_shape[i] = min(dim, self.min_shape[i])
                self.max_shape[i] = max(dim, self.max_shape[i])

    def __getstate__(self) -> Dict[str, Any]:
        d = super().__getstate__()

        for key in self._required_meta_keys:
            d[key] = getattr(self, key)

        return d

    def __setstate__(self, state: Dict[str, Any]):
        if "chunk_compression" not in state:
            state["chunk_compression"] = None  # Backward compatibility
        super().__setstate__(state)
        self._required_meta_keys = tuple(state.keys())

    @property
    def nbytes(self):
        # TODO: optimize this
        return len(self.tobytes())

    def __str__(self):
        return str(self.__getstate__())


def _required_meta_from_htype(htype: str) -> dict:
    """Gets a dictionary with all required meta information to define a tensor."""

    _validate_htype_exists(htype)
    defaults = HTYPE_CONFIGURATIONS[htype]

    required_meta = {
        "htype": htype,
        "min_shape": [],
        "max_shape": [],
        "length": 0,
        **defaults,
    }

    return required_meta


def _validate_htype_overwrites(htype: str, htype_overwrite: dict):
    """Raises errors if `htype_overwrite` has invalid keys or was missing required values."""

    defaults = HTYPE_CONFIGURATIONS[htype]

    for key, value in htype_overwrite.items():
        if key not in defaults:
            raise TensorMetaInvalidHtypeOverwriteKey(htype, key, list(defaults.keys()))

        if value == UNSPECIFIED:
            if defaults[key] == REQUIRE_USER_SPECIFICATION:
                raise TensorMetaMissingRequiredValue(htype, key)

    if (
        htype == "image"
        and htype_overwrite["chunk_compression"] == UNSPECIFIED
        and htype_overwrite["sample_compression"] == UNSPECIFIED
    ):
        raise TensorMetaMissingRequiredValue(
            htype, ["chunk_compression", "sample_compression"]  # type: ignore
        )

    if htype in ("json", "list", "text"):
        compr = htype_overwrite["chunk_compression"]
        if compr in (None, UNSPECIFIED):
            compr = htype_overwrite["sample_compression"]
        if compr not in (None, UNSPECIFIED):
            if get_compression_type(compr) != BYTE_COMPRESSION:
                raise UnsupportedCompressionError(compr, htype)
    elif htype == "audio":
        if htype_overwrite["chunk_compression"] not in [UNSPECIFIED, None]:
            raise UnsupportedCompressionError("Chunk compression", htype=htype)
        elif htype_overwrite["sample_compression"] == UNSPECIFIED:
            raise TensorMetaMissingRequiredValue(
                htype, "sample_compression"  # type: ignore
            )
        elif get_compression_type(htype_overwrite["sample_compression"]) not in (
            None,
            AUDIO_COMPRESSION,
        ):
            raise UnsupportedCompressionError(
                htype_overwrite["sample_compression"], htype="audio"
            )

    if htype == "video":
        if htype_overwrite["chunk_compression"] not in [UNSPECIFIED, None]:
            raise UnsupportedCompressionError("Chunk compression", htype=htype)
        elif htype_overwrite["sample_compression"] == UNSPECIFIED:
            raise TensorMetaMissingRequiredValue(
                htype, "sample compression"  # type: ignore
            )
        elif get_compression_type(htype_overwrite["sample_compression"]) not in (
            None,
            VIDEO_COMPRESSION,
        ):
            raise UnsupportedCompressionError(
                htype_overwrite["sample_compression"], htype="video"
            )


def _replace_unspecified_values(htype: str, htype_overwrite: dict):
    """Replaces `UNSPECIFIED` values in `htype_overwrite` with the `htype`'s defaults."""

    defaults = HTYPE_CONFIGURATIONS[htype]

    for k, v in htype_overwrite.items():
        if v == UNSPECIFIED:
            htype_overwrite[k] = defaults[k]

    if htype in ("json", "list", "text") and not htype_overwrite["dtype"]:
        htype_overwrite["dtype"] = HTYPE_CONFIGURATIONS[htype]["dtype"]


def _validate_required_htype_overwrites(htype: str, htype_overwrite: dict):
    """Raises errors if `htype_overwrite` has invalid values."""

    sample_compression = htype_overwrite["sample_compression"]
    sample_compression = COMPRESSION_ALIASES.get(sample_compression, sample_compression)
    if sample_compression not in hub.compressions:
        raise UnsupportedCompressionError(sample_compression)

    chunk_compression = htype_overwrite["chunk_compression"]
    chunk_compression = COMPRESSION_ALIASES.get(chunk_compression, chunk_compression)
    if chunk_compression not in hub.compressions:
        raise UnsupportedCompressionError(chunk_compression)

    if sample_compression and chunk_compression:
        raise TensorMetaMutuallyExclusiveKeysError(
            custom_message="Specifying both sample-wise and chunk-wise compressions for the same tensor is not yet supported."
        )

    if htype_overwrite["dtype"] is not None:
        if htype in ("json", "list"):
            validate_json_schema(htype_overwrite["dtype"])
        else:
            _raise_if_condition(
                "dtype",
                htype_overwrite,
                lambda dtype: not _is_dtype_supported_by_numpy(dtype),
                "Datatype must be supported by numpy. Can be an `str`, `np.dtype`, or normal python type (like `bool`, `float`, `int`, etc.). List of available numpy dtypes found here: https://numpy.org/doc/stable/user/basics.types.html",
            )

    if htype == "text":
        if htype_overwrite["dtype"] not in (str, "str"):
            raise TensorMetaInvalidHtypeOverwriteValue(
                "dtype",
                htype_overwrite["dtype"],
                "dtype for tensors with text htype should always be `str`",
            )


def _format_values(htype: str, htype_overwrite: dict):
    """Replaces values in `htype_overwrite` with consistent types/formats."""

    dtype = htype_overwrite["dtype"]
    if dtype is not None:
        if htype in ("json", "list"):
            if getattr(dtype, "__module__", None) == "typing":
                htype_overwrite["dtype"] = str(dtype)
        else:
            htype_overwrite["dtype"] = np.dtype(htype_overwrite["dtype"]).name

    for key, value in COMPRESSION_ALIASES.items():
        if htype_overwrite.get("sample_compression") == key:
            htype_overwrite["sample_compression"] = value
        if htype_overwrite.get("chunk_compression") == key:
            htype_overwrite["chunk_compression"] = value


def _validate_htype_exists(htype: str):
    """Raises errors if given an unrecognized htype."""
    if htype not in HTYPE_CONFIGURATIONS:
        raise TensorMetaInvalidHtype(htype, list(HTYPE_CONFIGURATIONS.keys()))


def _raise_if_condition(
    key: str, meta: dict, condition: Callable[[Any], bool], explanation: str = ""
):
    v = meta[key]
    if condition(v):
        raise TensorMetaInvalidHtypeOverwriteValue(key, v, explanation)


def _is_dtype_supported_by_numpy(dtype: str) -> bool:
    try:
        np.dtype(dtype)
        return True
    except:
        return False



from hub.util.exceptions import InvalidHubPathException


def check_hub_path(path):
    """Checks whether tag is in the format hub://username/datasetname."""
    tag = path[6:]
    if len(tag.split("/")) != 2:
        raise InvalidHubPathException(path)



from typing import Sequence
import numpy as np
from math import floor


def split(ds, values: Sequence[float] = [0.7, 0.2, 0.1]):
    """Splits a Dataset into multiple datasets with the provided ratio of entries.
    Returns a list of datasets with length equal to the number of givens.
    For small datasets or many partitions, some returns may be empty.

    Args:
        ds: The Dataset object from which to construct the splits.
            If already indexed, the splits will be based off that index.
        values: The proportions for the split. Should each sum to one.
            Defaults to [0.7, 0.2, 0.1] for a 70% 20% 10% split

    Returns:
        List of Datasets, one for each float in the given values.

    Raises:
        ValueError: The values must sum to 1.
    """

    if not np.isclose(sum(values), 1.0):
        raise ValueError("Given proportions must sum to 1.")

    count = 0
    length = len(ds)
    partitions = []
    for value in values[:-1]:
        amount = floor(length * value)
        partitions.append(ds[count : count + amount])
        count += amount
    partitions.append(ds[count:])

    return partitions



import pickle
import posixpath
import warnings
from typing import Any, Callable, Dict, List, Optional, Sequence, Tuple, Union

import hub
import numpy as np
from hub.api.info import load_info
from hub.client.log import logger
from hub.constants import FIRST_COMMIT_ID
from hub.core.fast_forwarding import ffw_dataset_meta
from hub.core.index import Index
from hub.core.lock import lock, unlock
from hub.core.meta.dataset_meta import DatasetMeta
from hub.core.storage import LRUCache, S3Provider
from hub.core.tensor import Tensor, create_tensor
from hub.core.version_control.commit_node import CommitNode  # type: ignore
from hub.htype import DEFAULT_HTYPE, HTYPE_CONFIGURATIONS, UNSPECIFIED
from hub.integrations import dataset_to_tensorflow
from hub.util.bugout_reporter import hub_reporter
from hub.util.exceptions import (
    CouldNotCreateNewDatasetException,
    InvalidKeyTypeError,
    InvalidTensorGroupNameError,
    InvalidTensorNameError,
    LockedException,
    MemoryDatasetCanNotBePickledError,
    PathNotEmptyException,
    TensorAlreadyExistsError,
    TensorDoesNotExistError,
    TensorGroupAlreadyExistsError,
)
from hub.util.keys import (
    dataset_exists,
    get_dataset_info_key,
    get_dataset_meta_key,
    get_version_control_info_key,
    tensor_exists,
)
from hub.util.path import get_path_from_storage
from hub.util.remove_cache import get_base_storage
from hub.util.version_control import (
    auto_checkout,
    checkout,
    commit,
    commit_has_data,
    load_meta,
)
from tqdm import tqdm  # type: ignore


class Dataset:
    def __init__(
        self,
        storage: LRUCache,
        index: Optional[Index] = None,
        group_index: str = "",
        read_only: bool = False,
        public: Optional[bool] = False,
        token: Optional[str] = None,
        verbose: bool = True,
        version_state: Optional[Dict[str, Any]] = None,
        path: Optional[str] = None,
        **kwargs,
    ):
        """Initializes a new or existing dataset.

        Args:
            storage (LRUCache): The storage provider used to access the dataset.
            index (Index, optional): The Index object restricting the view of this dataset's tensors.
            group_index (str): Name of the group this dataset instance represents.
            read_only (bool): Opens dataset in read only mode if this is passed as True. Defaults to False.
                Datasets stored on Hub cloud that your account does not have write access to will automatically open in read mode.
            public (bool, optional): Applied only if storage is Hub cloud storage and a new Dataset is being created. Defines if the dataset will have public access.
            token (str, optional): Activeloop token, used for fetching credentials for Hub datasets. This is optional, tokens are normally autogenerated.
            verbose (bool): If True, logs will be printed. Defaults to True.
            version_state (Dict[str, Any], optional): The version state of the dataset, includes commit_id, commit_node, branch, branch_commit_map and commit_node_map.
            **kwargs: Passing subclass variables through without errors.
            path: The path to the dataset.


        Raises:
            ValueError: If an existing local path is given, it must be a directory.
            ImproperDatasetInitialization: Exactly one argument out of 'path' and 'storage' needs to be specified.
                This is raised if none of them are specified or more than one are specifed.
            InvalidHubPathException: If a Hub cloud path (path starting with hub://) is specified and it isn't of the form hub://username/datasetname.
            AuthorizationException: If a Hub cloud path (path starting with hub://) is specified and the user doesn't have access to the dataset.
            PathNotEmptyException: If the path to the dataset doesn't contain a Hub dataset and is also not empty.
        """
        # uniquely identifies dataset
        self.path = path or get_path_from_storage(storage)
        self.storage = storage
        self._read_only = read_only
        base_storage = get_base_storage(storage)
        if (
            not read_only and index is None and isinstance(base_storage, S3Provider)
        ):  # Dataset locking only for S3 datasets
            try:
                lock(base_storage, callback=lambda: self._lock_lost_handler)
            except LockedException:
                self.read_only = True
                warnings.warn(
                    "Opening dataset in read only mode as another machine has locked it for writing."
                )

        self.index: Index = index or Index()
        self.group_index = group_index
        self._token = token
        self.public = public
        self.verbose = verbose
        self.version_state: Dict[str, Any] = version_state or {}
        self._info = None
        self._set_derived_attributes()

    def _lock_lost_handler(self):
        """This is called when lock is acquired but lost later on due to slow update."""
        self.read_only = True
        warnings.warn(
            "Unable to update dataset lock as another machine has locked it for writing. Switching to read only mode."
        )

    def __enter__(self):
        self.storage.autoflush = False
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.storage.autoflush = True
        self.flush()

    @property
    def num_samples(self) -> int:
        """Returns the length of the smallest tensor.
        Ignores any applied indexing and returns the total length.
        """
        return min(map(len, self.version_state["full_tensors"].values()), default=0)

    @property
    def meta(self) -> DatasetMeta:
        """Returns the metadata of the dataset."""
        return self.version_state["meta"]

    def __len__(self):
        """Returns the length of the smallest tensor"""
        tensor_lengths = [len(tensor) for tensor in self.tensors.values()]
        return min(tensor_lengths, default=0)

    def __getstate__(self) -> Dict[str, Any]:
        """Returns a dict that can be pickled and used to restore this dataset.

        Note:
            Pickling a dataset does not copy the dataset, it only saves attributes that can be used to restore the dataset.
            If you pickle a local dataset and try to access it on a machine that does not have the data present, the dataset will not work.
        """
        if self.path.startswith("mem://"):
            raise MemoryDatasetCanNotBePickledError
        return {
            "path": self.path,
            "_read_only": self.read_only,
            "index": self.index,
            "group_index": self.group_index,
            "public": self.public,
            "storage": self.storage,
            "_token": self.token,
            "verbose": self.verbose,
            "version_state": self.version_state,
        }

    def __setstate__(self, state: Dict[str, Any]):
        """Restores dataset from a pickled state.

        Args:
            state (dict): The pickled state used to restore the dataset.
        """
        self.__dict__.update(state)
        self._info = None
        self._set_derived_attributes()

    def __getitem__(
        self,
        item: Union[
            str, int, slice, List[int], Tuple[Union[int, slice, Tuple[int]]], Index
        ],
    ):
        if isinstance(item, str):
            fullpath = posixpath.join(self.group_index, item)
            tensor = self._get_tensor_from_root(fullpath)
            if tensor is not None:
                return tensor[self.index]
            elif self._has_group_in_root(fullpath):
                return self.__class__(
                    storage=self.storage,
                    index=self.index,
                    group_index=posixpath.join(self.group_index, item),
                    read_only=self.read_only,
                    token=self._token,
                    verbose=False,
                    version_state=self.version_state,
                    path=self.path,
                )
            elif "/" in item:
                splt = posixpath.split(item)
                return self[splt[0]][splt[1]]
            else:
                raise TensorDoesNotExistError(item)
        elif isinstance(item, (int, slice, list, tuple, Index)):
            return self.__class__(
                storage=self.storage,
                index=self.index[item],
                group_index=self.group_index,
                read_only=self.read_only,
                token=self._token,
                verbose=False,
                version_state=self.version_state,
                path=self.path,
            )
        else:
            raise InvalidKeyTypeError(item)

    @hub_reporter.record_call
    def create_tensor(
        self,
        name: str,
        htype: str = DEFAULT_HTYPE,
        dtype: Union[str, np.dtype] = UNSPECIFIED,
        sample_compression: str = UNSPECIFIED,
        chunk_compression: str = UNSPECIFIED,
        **kwargs,
    ):
        """Creates a new tensor in the dataset.

        Args:
            name (str): The name of the tensor to be created.
            htype (str): The class of data for the tensor.
                The defaults for other parameters are determined in terms of this value.
                For example, `htype="image"` would have `dtype` default to `uint8`.
                These defaults can be overridden by explicitly passing any of the other parameters to this function.
                May also modify the defaults for other parameters.
            dtype (str): Optionally override this tensor's `dtype`. All subsequent samples are required to have this `dtype`.
            sample_compression (str): All samples will be compressed in the provided format. If `None`, samples are uncompressed.
            chunk_compression (str): All chunks will be compressed in the provided format. If `None`, chunks are uncompressed.
            **kwargs: `htype` defaults can be overridden by passing any of the compatible parameters.
                To see all `htype`s and their correspondent arguments, check out `hub/htypes.py`.

        Returns:
            The new tensor, which can also be accessed by `self[name]`.

        Raises:
            TensorAlreadyExistsError: Duplicate tensors are not allowed.
            TensorGroupAlreadyExistsError: Duplicate tensor groups are not allowed.
            InvalidTensorNameError: If `name` is in dataset attributes.
            NotImplementedError: If trying to override `chunk_compression`.
        """
        # if not the head node, checkout to an auto branch that is newly created
        auto_checkout(self.version_state, self.storage)
        name = name.strip("/")

        while "//" in name:
            name = name.replace("//", "/")

        full_path = posixpath.join(self.group_index, name)

        if tensor_exists(full_path, self.storage, self.version_state["commit_id"]):
            raise TensorAlreadyExistsError(name)

        if full_path in self._groups:
            raise TensorGroupAlreadyExistsError(name)

        if not name or name in dir(self):
            raise InvalidTensorNameError(name)

        if not self._is_root():
            return self.root.create_tensor(
                full_path, htype, dtype, sample_compression, chunk_compression, **kwargs
            )

        if "/" in name:
            self._create_group(posixpath.split(name)[0])

        # Seperate meta and info

        htype_config = HTYPE_CONFIGURATIONS[htype].copy()
        info_keys = htype_config.pop("_info", [])
        info_kwargs = {}
        meta_kwargs = {}
        for k, v in kwargs.items():
            if k in info_keys:
                info_kwargs[k] = v
            else:
                meta_kwargs[k] = v

        # Set defaults
        for k in info_keys:
            if k not in info_kwargs:
                info_kwargs[k] = htype_config[k]

        create_tensor(
            name,
            self.storage,
            htype=htype,
            dtype=dtype,
            sample_compression=sample_compression,
            chunk_compression=chunk_compression,
            version_state=self.version_state,
            **meta_kwargs,
        )
        self.version_state["meta"].tensors.append(name)
        ffw_dataset_meta(self.version_state["meta"])
        self.storage.maybe_flush()
        tensor = Tensor(name, self.storage, self.version_state)  # type: ignore

        self.version_state["full_tensors"][name] = tensor
        tensor.info.update(info_kwargs)
        return tensor

    @hub_reporter.record_call
    def create_tensor_like(self, name: str, source: "Tensor") -> "Tensor":
        """Copies the `source` tensor's meta information and creates a new tensor with it. No samples are copied, only the meta/info for the tensor is.

        Args:
            name (str): Name for the new tensor.
            source (Tensor): Tensor who's meta/info will be copied. May or may not be contained in the same dataset.

        Returns:
            Tensor: New Tensor object.
        """

        info = source.info.__getstate__().copy()
        meta = source.meta.__getstate__().copy()
        del meta["min_shape"]
        del meta["max_shape"]
        del meta["length"]
        del meta["version"]

        destination_tensor = self.create_tensor(
            name,
            **meta,
        )
        destination_tensor.info.update(info)

        return destination_tensor

    __getattr__ = __getitem__

    def __setattr__(self, name: str, value):
        if isinstance(value, (np.ndarray, np.generic)):
            raise TypeError(
                "Setting tensor attributes directly is not supported. To add a tensor, use the `create_tensor` method."
                + "To add data to a tensor, use the `append` and `extend` methods."
            )
        else:
            return super().__setattr__(name, value)

    def __iter__(self):
        for i in range(len(self)):
            yield self[i]

    def _load_version_info(self):
        """Loads data from version_control_file otherwise assume it doesn't exist and load all empty"""
        branch = "main"
        version_state = {"branch": branch}
        try:
            version_info = pickle.loads(self.storage[get_version_control_info_key()])
            version_state["branch_commit_map"] = version_info["branch_commit_map"]
            version_state["commit_node_map"] = version_info["commit_node_map"]
            commit_id = version_state["branch_commit_map"][branch]
            version_state["commit_id"] = commit_id
            version_state["commit_node"] = version_state["commit_node_map"][commit_id]
        except Exception:
            version_state["branch_commit_map"] = {}
            version_state["commit_node_map"] = {}
            # used to identify that this is the first commit so its data will not be in similar directory structure to the rest
            commit_id = FIRST_COMMIT_ID
            commit_node = CommitNode(branch, commit_id)
            version_state["commit_id"] = commit_id
            version_state["commit_node"] = commit_node
            version_state["branch_commit_map"][branch] = commit_id
            version_state["commit_node_map"][commit_id] = commit_node
        version_state["full_tensors"] = {}  # keeps track of the full unindexed tensors
        self.version_state = version_state

    def commit(self, message: Optional[str] = None) -> None:
        """Stores a snapshot of the current state of the dataset.
        Note: Commiting from a non-head node in any branch, will lead to an auto checkout to a new branch.
        This same behaviour will happen if new samples are added or existing samples are updated from a non-head node.

        Args:
            message (str, optional): Used to describe the commit.

        Returns:
            str: the commit id of the stored commit that can be used to access the snapshot.
        """
        commit_id = self.version_state["commit_id"]
        commit(self.version_state, self.storage, message)

        # do not store commit message
        hub_reporter.feature_report(
            feature_name="commit",
            parameters={},
        )

        return commit_id

    def checkout(self, address: str, create: bool = False) -> str:
        """Checks out to a specific commit_id or branch. If create = True, creates a new branch with name as address.
        Note: Checkout from a head node in any branch that contains uncommitted data will lead to an auto commit before the checkout.

        Args:
            address (str): The commit_id or branch to checkout to.
            create (bool): If True, creates a new branch with name as address.

        Returns:
            str: The commit_id of the dataset after checkout.
        """
        checkout(self.version_state, self.storage, address, create)

        # do not store address
        hub_reporter.feature_report(
            feature_name="checkout",
            parameters={"Create": str(create)},
        )

        return self.version_state["commit_id"]

    def log(self):
        """Displays the details of all the past commits."""
        commit_node = self.version_state["commit_node"]
        logger.info("---------------\nHub Version Log\n---------------\n")
        logger.info(f"Current Branch: {self.version_state['branch']}")
        if not commit_node.children and commit_has_data(
            self.version_state, self.storage
        ):
            logger.info("** There are uncommitted changes on this branch.\n")
        else:
            logger.info("\n")
        while commit_node:
            if commit_node.commit_time is not None:
                logger.info(f"{commit_node}\n")
            commit_node = commit_node.parent

    def _populate_meta(self):
        """Populates the meta information for the dataset."""

        if dataset_exists(self.storage):
            if self.verbose:
                logger.info(f"{self.path} loaded successfully.")
            load_meta(self.storage, self.version_state)

        elif not self.storage.empty():
            # dataset does not exist, but the path was not empty
            raise PathNotEmptyException

        else:
            if self.read_only:
                # cannot create a new dataset when in read_only mode.
                raise CouldNotCreateNewDatasetException(self.path)
            meta_key = get_dataset_meta_key(self.version_state["commit_id"])
            self.version_state["meta"] = DatasetMeta()
            self.storage[meta_key] = self.version_state["meta"]
            self.flush()
            self._register_dataset()

    def _register_dataset(self):
        # overridden in HubCloudDataset

        pass

    @property
    def read_only(self):
        return self._read_only

    @read_only.setter
    def read_only(self, value: bool):
        if value:
            self.storage.enable_readonly()
        else:
            self.storage.disable_readonly()
        self._read_only = value

    @hub_reporter.record_call
    def pytorch(
        self,
        transform: Optional[Callable] = None,
        tensors: Optional[Sequence[str]] = None,
        num_workers: int = 1,
        batch_size: int = 1,
        drop_last: bool = False,
        collate_fn: Optional[Callable] = None,
        pin_memory: bool = False,
        shuffle: bool = False,
        buffer_size: int = 2048,
        use_local_cache: bool = False,
        use_progress_bar: bool = False,
    ):
        """Converts the dataset into a pytorch Dataloader.

        Note:
            Pytorch does not support uint16, uint32, uint64 dtypes. These are implicitly type casted to int32, int64 and int64 respectively.
            This spins up it's own workers to fetch data.

        Args:
            transform (Callable, optional) : Transformation function to be applied to each sample.
            tensors (List, optional): Optionally provide a list of tensor names in the ordering that your training script expects. For example, if you have a dataset that has "image" and "label" tensors, if `tensors=["image", "label"]`, your training script should expect each batch will be provided as a tuple of (image, label).
            num_workers (int): The number of workers to use for fetching data in parallel.
            batch_size (int): Number of samples per batch to load. Default value is 1.
            drop_last (bool): Set to True to drop the last incomplete batch, if the dataset size is not divisible by the batch size.
                If False and the size of dataset is not divisible by the batch size, then the last batch will be smaller. Default value is False.
                Read torch.utils.data.DataLoader docs for more details.
            collate_fn (Callable, optional): merges a list of samples to form a mini-batch of Tensor(s). Used when using batched loading from a map-style dataset.
                Read torch.utils.data.DataLoader docs for more details.
            pin_memory (bool): If True, the data loader will copy Tensors into CUDA pinned memory before returning them. Default value is False.
                Read torch.utils.data.DataLoader docs for more details.
            shuffle (bool): If True, the data loader will shuffle the data indices. Default value is False.
            buffer_size (int): The size of the buffer used to prefetch/shuffle in MB. The buffer uses shared memory under the hood. Default value is 2 GB. Increasing the buffer_size will increase the extent of shuffling.
            use_local_cache (bool): If True, the data loader will use a local cache to store data. This is useful when the dataset can fit on the machine and we don't want to fetch the data multiple times for each iteration. Default value is False.
            use_progress_bar (bool): If True, tqdm will be wrapped around the returned dataloader. Default value is True.

        Returns:
            A torch.utils.data.DataLoader object.
        """
        from hub.integrations import dataset_to_pytorch as to_pytorch

        dataloader = to_pytorch(
            self,
            transform=transform,
            tensors=tensors,
            num_workers=num_workers,
            batch_size=batch_size,
            drop_last=drop_last,
            collate_fn=collate_fn,
            pin_memory=pin_memory,
            shuffle=shuffle,
            buffer_size=buffer_size,
            use_local_cache=use_local_cache,
        )

        if use_progress_bar:
            dataloader = tqdm(dataloader, desc=self.path, total=len(self) // batch_size)

        return dataloader

    def _get_total_meta(self):
        """Returns tensor metas all together"""
        return {
            tensor_key: tensor_value.meta
            for tensor_key, tensor_value in self.version_state["full_tensors"].items()
        }

    def _set_derived_attributes(self):
        """Sets derived attributes during init and unpickling."""
        if self.index.is_trivial() and self._is_root():
            self.storage.autoflush = True

        if not self.version_state:
            self._load_version_info()

        self._populate_meta()  # TODO: use the same scheme as `load_info`
        self.read_only = self._read_only  # TODO: weird fix for dataset unpickling
        self.index.validate(self.num_samples)

    @property
    def info(self):
        if self._info is None:
            self._info = load_info(get_dataset_info_key(self.version_state["commit_id"]), self.storage, self.version_state)  # type: ignore
        return self._info

    @hub_reporter.record_call
    def tensorflow(self):
        """Converts the dataset into a tensorflow compatible format.

        See:
            https://www.tensorflow.org/api_docs/python/tf/data/Dataset

        Returns:
            tf.data.Dataset object that can be used for tensorflow training.
        """
        return dataset_to_tensorflow(self)

    def flush(self):
        """Necessary operation after writes if caches are being used.
        Writes all the dirty data from the cache layers (if any) to the underlying storage.
        Here dirty data corresponds to data that has been changed/assigned and but hasn't yet been sent to the
        underlying storage.
        """
        self.storage.flush()

    def clear_cache(self):
        """Flushes (see Dataset.flush documentation) the contents of the cache layers (if any) and then deletes contents
         of all the layers of it.
        This doesn't delete data from the actual storage.
        This is useful if you have multiple datasets with memory caches open, taking up too much RAM.
        Also useful when local cache is no longer needed for certain datasets and is taking up storage space.
        """
        if hasattr(self.storage, "clear_cache"):
            self.storage.clear_cache()

    def size_approx(self):
        """Estimates the size in bytes of the dataset.
        Includes only content, so will generally return an under-estimate.
        """
        tensors = self.version_state["full_tensors"].values()
        chunk_engines = [tensor.chunk_engine for tensor in tensors]
        size = sum(c.num_chunks * c.min_chunk_size for c in chunk_engines)
        return size

    @hub_reporter.record_call
    def delete(self, large_ok=False):
        """Deletes the entire dataset from the cache layers (if any) and the underlying storage.
        This is an IRREVERSIBLE operation. Data once deleted can not be recovered.

        Args:
            large_ok (bool): Delete datasets larger than 1GB. Disabled by default.
        """

        if not large_ok:
            size = self.size_approx()
            if size > hub.constants.DELETE_SAFETY_SIZE:
                logger.info(
                    f"Hub Dataset {self.path} was too large to delete. Try again with large_ok=True."
                )
                return

        unlock(self.storage)
        self.storage.clear()

    def __str__(self):
        path_str = ""
        if self.path:
            path_str = f"path='{self.path}', "

        mode_str = ""
        if self.read_only:
            mode_str = f"read_only=True, "

        index_str = f"index={self.index}, "
        if self.index.is_trivial():
            index_str = ""

        group_index_str = (
            f"group_index='{self.group_index}', " if self.group_index else ""
        )

        return f"Dataset({path_str}{mode_str}{index_str}{group_index_str}tensors={self.version_state['meta'].tensors})"

    __repr__ = __str__

    def _get_tensor_from_root(self, name: str) -> Optional[Tensor]:
        """Gets a tensor from the root dataset.
        Acesses storage only for the first call.
        """
        ret = self.version_state["full_tensors"].get(name)
        if ret is None:
            load_meta(self.storage, self.version_state)
            ret = self.version_state["full_tensors"].get(name)
        return ret

    def _has_group_in_root(self, name: str) -> bool:
        """Checks if a group exists in the root dataset.
        This is faster than checking `if group in self._groups:`
        """
        if name in self.version_state["meta"].groups:
            return True
        load_meta(self.storage, self.version_state)
        return name in self.version_state["meta"].groups

    @property
    def token(self):
        """Get attached token of the dataset"""

        return self._token

    @property
    def _ungrouped_tensors(self) -> Dict[str, Tensor]:
        """Top level tensors in this group that do not belong to any sub groups"""
        return {
            posixpath.basename(k): v
            for k, v in self.version_state["full_tensors"].items()
            if posixpath.dirname(k) == self.group_index
        }

    @property
    def _all_tensors_filtered(self) -> List[str]:
        """Names of all tensors belonging to this group, including those within sub groups"""
        load_meta(self.storage, self.version_state)
        return [
            posixpath.relpath(t, self.group_index)
            for t in self.version_state["full_tensors"]
            if not self.group_index or t.startswith(self.group_index + "/")
        ]

    @property
    def tensors(self) -> Dict[str, Tensor]:
        """All tensors belonging to this group, including those within sub groups. Always returns the sliced tensors."""
        return {
            t: self.version_state["full_tensors"][posixpath.join(self.group_index, t)][
                self.index
            ]
            for t in self._all_tensors_filtered
        }

    @property
    def _groups(self) -> List[str]:
        """Names of all groups in the root dataset"""
        meta_key = get_dataset_meta_key(self.version_state["commit_id"])
        return self.storage.get_cachable(meta_key, DatasetMeta).groups  # type: ignore

    @property
    def _groups_filtered(self) -> List[str]:
        """Names of all sub groups in this group"""
        groups_filtered = []
        for g in self._groups:
            dirname, basename = posixpath.split(g)
            if dirname == self.group_index:
                groups_filtered.append(basename)
        return groups_filtered

    @property
    def groups(self) -> Dict[str, "Dataset"]:
        """All sub groups in this group"""
        return {g: self[g] for g in self._groups_filtered}

    @property
    def commit_id(self) -> str:
        """The current commit_id of the dataset."""
        return self.version_state["commit_id"]

    @property
    def branch(self) -> str:
        """The current branch of the dataset"""
        return self.version_state["branch"]

    def _is_root(self) -> bool:
        return not self.group_index

    @property
    def parent(self):
        """Returns the parent of this group. Returns None if this is the root dataset"""
        if self._is_root():
            return None
        autoflush = self.storage.autoflush
        ds = self.__class__(
            storage=self.storage,
            index=self.index,
            group_index=posixpath.dirname(self.group_index),
            read_only=self.read_only,
            public=self.public,
            token=self._token,
            verbose=self.verbose,
            path=self.path,
        )
        self.storage.autoflush = autoflush
        return ds

    @property
    def root(self):
        if self._is_root():
            return self
        autoflush = self.storage.autoflush
        ds = self.__class__(
            storage=self.storage,
            index=self.index,
            group_index="",
            read_only=self.read_only,
            public=self.public,
            token=self._token,
            verbose=self.verbose,
            path=self.path,
        )
        self.storage.autoflush = autoflush
        return ds

    def _create_group(self, name: str) -> "Dataset":
        """Internal method used by `create_group` and `create_tensor`."""
        meta_key = get_dataset_meta_key(self.version_state["commit_id"])
        meta = self.storage.get_cachable(meta_key, DatasetMeta)
        groups = meta.groups
        if not name or name in dir(self):
            raise InvalidTensorGroupNameError(name)
        fullname = name
        while name:
            if name in self.version_state["full_tensors"]:
                raise TensorAlreadyExistsError(name)
            groups.append(name)
            name, _ = posixpath.split(name)
        meta.groups = list(set(groups))
        self.storage[meta_key] = meta
        self.storage.maybe_flush()
        return self[fullname]

    def create_group(self, name: str) -> "Dataset":
        """Creates a tensor group. Intermediate groups in the path are also created."""
        if not self._is_root():
            return self.root.create_group(posixpath.join(self.group_index, name))
        name = name.strip("/")
        while "//" in name:
            name = name.replace("//", "/")
        if name in self._groups:
            raise TensorGroupAlreadyExistsError(name)
        return self._create_group(name)

    # the below methods are used by cloudpickle dumps
    def __origin__(self):
        return None

    def __values__(self):
        return None

    def __type__(self):
        return None

    def __union_params__(self):
        return None

    def __tuple_params__(self):
        return None

    def __result__(self):
        return None

    def __args__(self):
        return None



from hub.util.iterable_ordered_dict import IterableOrderedDict as OrderedDict
import pickle


def test_ordereddict():
    T = OrderedDict
    t1 = T()
    t1["b"] = 2
    t1["a"] = 1
    t1["d"] = 4
    t1["c"] = 3

    t2 = T(b=2, a=1, d=4, c=3)

    assert t1 == t2
    assert t1["b"] == 2
    assert t1["c"] == 3
    assert t1["d"] == 4
    assert t2["a"] == 1
    assert t2["b"] == 2
    assert t2["c"] == 3
    assert t2["d"] == 4
    assert t1 == t2
    assert len(t1) == len(t2) == 4
    assert list(t1.keys()) == list(t2.keys()) == ["b", "a", "d", "c"]
    b1, a1, d1, c1 = t1
    b2, a2, d2, c2 = t2
    assert a1 == a2 == 1
    assert b1 == b2 == 2
    assert c1 == c2 == 3
    assert d1 == d2 == 4

    pickled = pickle.dumps(t2)
    t2 = pickle.loads(pickled)

    assert t1 == t2
    assert t1["b"] == 2
    assert t1["c"] == 3
    assert t1["d"] == 4
    assert t2["a"] == 1
    assert t2["b"] == 2
    assert t2["c"] == 3
    assert t2["d"] == 4
    assert t1 == t2
    assert len(t1) == len(t2) == 4
    assert list(t1.keys()) == list(t2.keys()) == ["b", "a", "d", "c"]
    b1, a1, d1, c1 = t1
    b2, a2, d2, c2 = t2
    assert a1 == a2 == 1
    assert b1 == b2 == 2
    assert c1 == c2 == 3
    assert d1 == d2 == 4



from typing import Callable, Optional, Sequence
from hub.util.dataset import try_flushing
from hub.constants import MB
from .common import convert_fn as default_convert_fn, collate_fn as default_collate_fn


def create_dataloader_nesteddataloader(
    dataset,
    tensors,
    use_local_cache,
    transform,
    num_workers,
    buffer_size,
    batch_size,
    collate_fn,
    pin_memory,
    drop_last,
):
    import torch
    import torch.utils.data
    from hub.integrations.pytorch.dataset import SubIterableDataset

    return torch.utils.data.DataLoader(
        # this data set is more efficient also shuffles
        # using threads race conditions as source of entropy
        SubIterableDataset(
            dataset,
            tensors=tensors,
            use_local_cache=use_local_cache,
            transform=transform,
            num_workers=num_workers,
            buffer_size=buffer_size,
            batch_size=batch_size,
            collate_fn=collate_fn,
        ),
        batch_size=batch_size,
        collate_fn=collate_fn,
        pin_memory=pin_memory,
        drop_last=drop_last,
    )


def create_dataloader_shufflingdataloader(
    dataset,
    tensors,
    use_local_cache,
    transform,
    num_workers,
    buffer_size,
    batch_size,
    collate_fn,
    pin_memory,
    drop_last,
):
    import torch
    import torch.utils.data
    from hub.integrations.pytorch.dataset import ShufflingIterableDataset

    return torch.utils.data.DataLoader(
        # this data set is more efficient also shuffles
        # using threads race conditions as source of entropy
        ShufflingIterableDataset(
            dataset,
            tensors=tensors,
            use_local_cache=use_local_cache,
            transform=transform,
            num_workers=num_workers,
            buffer_size=buffer_size,
        ),
        batch_size=batch_size,
        collate_fn=collate_fn,
        pin_memory=pin_memory,
        drop_last=drop_last,
    )


create_dataloader = create_dataloader_nesteddataloader


def dataset_to_pytorch(
    dataset,
    num_workers: int,
    batch_size: int,
    drop_last: bool,
    collate_fn: Optional[Callable],
    pin_memory: bool,
    shuffle: bool,
    buffer_size: int,
    use_local_cache: bool,
    transform: Optional[Callable] = None,
    tensors: Optional[Sequence[str]] = None,
):

    import torch
    from hub.integrations.pytorch.dataset import TorchDataset

    try_flushing(dataset)

    torch.multiprocessing.set_sharing_strategy("file_system")

    if collate_fn is None:
        collate_fn = default_convert_fn if batch_size is None else default_collate_fn

    if shuffle and num_workers > 0:
        return create_dataloader(
            dataset,
            tensors,
            use_local_cache,
            transform,
            num_workers,
            buffer_size,
            batch_size,
            collate_fn,
            pin_memory,
            drop_last,
        )
    else:
        return torch.utils.data.DataLoader(
            TorchDataset(
                dataset,
                tensors=tensors,
                use_local_cache=use_local_cache,
                transform=transform,
                num_workers=num_workers,
                shuffle=False,
                buffer_size=buffer_size,
            ),
            batch_size=batch_size,
            collate_fn=collate_fn,
            pin_memory=pin_memory,
            num_workers=num_workers,
            drop_last=drop_last,
        )



# type: ignore
from hub.core.compression import (
    compress_array,
    compress_bytes,
    decompress_array,
    verify_compressed_file,
    read_meta_from_compressed_file,
    get_compression,
    _to_hub_mkv,
)
from hub.compression import (
    get_compression_type,
    AUDIO_COMPRESSION,
    IMAGE_COMPRESSION,
    VIDEO_COMPRESSION,
)
from hub.compression import get_compression_type, AUDIO_COMPRESSION, IMAGE_COMPRESSION
from hub.util.exceptions import CorruptedSampleError
import numpy as np
from typing import List, Optional, Tuple, Union

from PIL import Image  # type: ignore
from io import BytesIO


class Sample:
    path: Optional[str]

    def __init__(
        self,
        path: str = None,
        array: np.ndarray = None,
        buffer: Union[bytes, memoryview] = None,
        compression: str = None,
        verify: bool = False,
    ):
        """Represents a single sample for a tensor. Provides all important meta information in one place.

        Note:
            If `self.is_lazy` is True, this `Sample` doesn't actually have any data loaded. To read this data,
                simply try to read it into a numpy array (`sample.array`)

        Args:
            path (str): Path to a sample stored on the local file system that represents a single sample. If `path` is provided, `array` should not be.
                Implicitly makes `self.is_lazy == True`.
            array (np.ndarray): Array that represents a single sample. If `array` is provided, `path` should not be. Implicitly makes `self.is_lazy == False`.
            buffer: (bytes): Byte buffer that represents a single sample. If compressed, `compression` argument should be provided.
            compression (str): Specify in case of byte buffer.
            verify (bool): If a path is provided, verifies the sample if True.

        Raises:
            ValueError: Cannot create a sample from both a `path` and `array`.
        """

        if not any((path, array, buffer)):
            raise ValueError("Must pass one of `path`, `array` or `buffer`.")

        self._compressed_bytes = {}
        self._uncompressed_bytes = None

        self._array = None
        self._typestr = None
        self._shape = None
        self.path = None
        self._buffer = None

        if path is not None:
            self.path = path
            self._compression = None
            self._verified = False
            self._verify = verify

        if array is not None:
            self._array = array
            self._shape = array.shape
            self._typestr = array.__array_interface__["typestr"]
            self._compression = None

        if buffer is not None:
            self._compression = compression
            self._buffer = buffer
            if compression is None:
                self._uncompressed_bytes = buffer
            else:
                self._compressed_bytes[compression] = buffer

    @property
    def buffer(self):
        if self._buffer is not None:
            return self._buffer
        return self.compressed_bytes(self.compression)

    @property
    def dtype(self):
        self._read_meta()
        return np.dtype(self._typestr).name

    @property
    def shape(self):
        self._read_meta()
        return self._shape

    @property
    def compression(self):
        if self._compression is None and self.path:
            self._read_meta()
        return self._compression

    def _read_meta(self, f=None):
        if self._shape is not None:
            return
        if f is None:
            f = self.path if self.path else self.compressed_bytes[self._compression]
        self._compression, self._shape, self._typestr = read_meta_from_compressed_file(
            f
        )

    @property
    def is_lazy(self) -> bool:
        return self._array is None

    @property
    def is_empty(self) -> bool:
        return 0 in self.shape

    def _recompress(self, buffer: bytes, compression: str) -> bytes:
        if get_compression_type(self._compression) != IMAGE_COMPRESSION:
            raise ValueError(
                "Recompression with different format is only supported for images."
            )
        img = Image.open(BytesIO(buffer))
        if img.mode == "1":
            self._uncompressed_bytes = img.tobytes("raw", "L")
        else:
            self._uncompressed_bytes = img.tobytes()
        return compress_array(self.array, compression)

    def compressed_bytes(self, compression: str) -> bytes:
        """Returns this sample as compressed bytes.

        Note:
            If this sample is pointing to a path and the requested `compression` is the same as it's stored in, the data is
                returned without re-compressing.

        Args:
            compression (str): `self.array` will be compressed into this format. If `compression is None`, return `self.uncompressed_bytes()`.

        Returns:
            bytes: Bytes for the compressed sample. Contains all metadata required to decompress within these bytes.

        Raises:
            ValueError: On recompression of unsupported formats.
        """

        if compression is None:
            return self.uncompressed_bytes()

        compressed_bytes = self._compressed_bytes.get(compression)
        if compressed_bytes is None:
            if self.path is not None:
                if self._compression is None:
                    self._compression = get_compression(path=self.path)
                if self._compression in (
                    "mp4",
                    "mkv",
                ):  # mp4 byte stream is not seekable, may not be able to extract duration from mkv byte stream
                    compressed_bytes = _to_hub_mkv(self.path)
                else:
                    with open(self.path, "rb") as f:
                        compressed_bytes = f.read()
                    if self._compression is None:
                        self._compression = get_compression(
                            header=compressed_bytes[:32]
                        )
                if self._compression == compression:
                    if self._verify:
                        self._shape, self._typestr = verify_compressed_file(
                            compressed_bytes, self._compression
                        )
                    else:
                        _, self._shape, self._typestr = read_meta_from_compressed_file(
                            compressed_bytes, compression=self._compression
                        )
                else:
                    compressed_bytes = self._recompress(compressed_bytes, compression)
            elif self._buffer is not None:
                compressed_bytes = self._recompress(self._buffer, compression)
            else:
                compressed_bytes = compress_array(self.array, compression)
            self._compressed_bytes[compression] = compressed_bytes
        return compressed_bytes

    def uncompressed_bytes(self) -> bytes:
        """Returns uncompressed bytes."""

        if self._uncompressed_bytes is None:
            if self.path is not None:
                compr = self._compression
                if compr is None:
                    compr = get_compression(path=self.path)
                if get_compression_type(compr) in (
                    AUDIO_COMPRESSION,
                    VIDEO_COMPRESSION,
                ):
                    self._compression = compr
                    if self._array is None:
                        self._array = decompress_array(self.path, compression=compr)
                    self._uncompressed_bytes = self._array.tobytes()
                else:
                    img = Image.open(self.path)
                    if img.mode == "1":
                        # Binary images need to be extended from bits to bytes
                        self._uncompressed_bytes = img.tobytes("raw", "L")
                    else:
                        self._uncompressed_bytes = img.tobytes()
            else:
                self._uncompressed_bytes = self._array.tobytes()

        return self._uncompressed_bytes

    @property
    def array(self) -> np.ndarray:

        if self._array is None:
            compr = self._compression
            if compr is None:
                compr = get_compression(path=self.path)
            if get_compression_type(compr) in (AUDIO_COMPRESSION, VIDEO_COMPRESSION):
                self._compression = compr
                array = decompress_array(self.path or self._buffer, compression=compr)
                if self._shape is None:
                    self._shape = array.shape
                    self._typestr = array.__array_interface__["typestr"]
                self._array = array
            else:
                self._read_meta()
                array_interface = {
                    "shape": self._shape,
                    "typestr": self._typestr,
                    "version": 3,
                    "data": self.uncompressed_bytes(),
                }

                class ArrayData:
                    __array_interface__ = array_interface

                self._array = np.array(ArrayData, None)
        return self._array

    def __str__(self):
        if self.is_lazy:
            return f"Sample(is_lazy=True, path={self.path})"

        return f"Sample(is_lazy=False, shape={self.shape}, compression='{self.compression}', dtype='{self.dtype}' path={self.path})"

    def __repr__(self):
        return str(self)

    def __array__(self):
        return self.array

    def __eq__(self, other):
        if self.path is not None and other.path is not None:
            return self.path == other.path
        return self.buffer == other.buffer


SampleValue = Union[np.ndarray, int, float, bool, Sample]



from hub.util.keys import get_dataset_meta_key, get_tensor_meta_key
from hub.core.storage.provider import StorageProvider
from hub.core.storage import LRUCache
import glob
import os


def is_hub_cloud_path(path: str):
    return path.startswith("hub://")


def get_path_from_storage(storage) -> str:
    """Extracts the underlying path from a given storage."""
    if isinstance(storage, LRUCache):
        return get_path_from_storage(storage.next_storage)
    elif isinstance(storage, StorageProvider):
        if hasattr(storage, "hub_path"):
            return storage.hub_path  # type: ignore
        return storage.root
    else:
        raise ValueError("Invalid storage type.")


def find_root(path: str) -> str:
    """Find the root of the dataset within the given path.

    Note:
        The "root" is defined as the subdirectory (within path) that has > 1 folder/file (if applicable).
        in other words, if there is a directory structure like:
        dataset -
            Images -
                class1 -
                    img.jpg
                    ...
                class2 -
                    img.jpg
                    ...
                ...

        root is "dataset/Images"

    Args:
        path (str): The local path to folder containing an unstructured dataset and of the form ./path/to/dataset or ~/path/to/dataset or path/to/dataset.

    Returns:
        str: Root path of the unstructured dataset.
    """

    subs = glob.glob(os.path.join(path, "*"))

    subs = [
        sub for sub in subs if os.path.isdir(sub)
    ]  # only keep directories (ignore files)

    if len(subs) == 1:
        return find_root(subs[0])

    return path



import hub
import math
from typing import List, Callable, Optional
from itertools import repeat
from hub.core.compute.provider import ComputeProvider
from hub.core.compute.thread import ThreadProvider
from hub.core.compute.process import ProcessProvider
from hub.core.compute.serial import SerialProvider
from hub.core.ipc import Server
from hub.util.bugout_reporter import hub_reporter
from hub.util.compute import get_compute_provider
from hub.util.remove_cache import get_base_storage, get_dataset_with_zero_size_cache
from hub.util.transform import (
    check_transform_data_in,
    check_transform_ds_out,
    get_pbar_description,
    store_data_slice,
)
from hub.util.encoder import merge_all_chunk_id_encoders, merge_all_tensor_metas
from hub.util.exceptions import (
    HubComposeEmptyListError,
    HubComposeIncompatibleFunction,
    TransformError,
)

from tqdm import tqdm  # type: ignore
import time
import threading
import sys


class TransformFunction:
    def __init__(self, func, args, kwargs):
        """Creates a TransformFunction object that can be evaluated using .eval or used as a part of a Pipeline."""
        self.func = func
        self.args = args
        self.kwargs = kwargs

    def eval(
        self,
        data_in,
        ds_out: hub.Dataset,
        num_workers: int = 0,
        scheduler: str = "threaded",
        progressbar: bool = True,
    ):
        """Evaluates the TransformFunction on data_in to produce an output dataset ds_out.

        Args:
            data_in: Input passed to the transform to generate output dataset. Should support __getitem__ and __len__. Can be a Hub dataset.
            ds_out (Dataset): The dataset object to which the transform will get written.
                Should have all keys being generated in output already present as tensors. It's initial state should be either:-
                - Empty i.e. all tensors have no samples. In this case all samples are added to the dataset.
                - All tensors are populated and have sampe length. In this case new samples are appended to the dataset.
            num_workers (int): The number of workers to use for performing the transform. Defaults to 0. When set to 0, it will always use serial processing, irrespective of the scheduler.
            scheduler (str): The scheduler to be used to compute the transformation. Supported values include: "serial", 'threaded', 'processed' and 'ray.
            progressbar (bool): Displays a progress bar if True (default).


        Raises:
            InvalidInputDataError: If data_in passed to transform is invalid. It should support __getitem__ and __len__ operations. Using scheduler other than "threaded" with hub dataset having base storage as memory as data_in will also raise this.
            InvalidOutputDatasetError: If all the tensors of ds_out passed to transform don't have the same length. Using scheduler other than "threaded" with hub dataset having base storage as memory as ds_out will also raise this.
            TensorMismatchError: If one or more of the outputs generated during transform contain different tensors than the ones present in 'ds_out' provided to transform.
            UnsupportedSchedulerError: If the scheduler passed is not recognized. Supported values include: "serial", 'threaded', 'processed' and 'ray'.
        """

        pipeline = Pipeline([self])
        pipeline.eval(data_in, ds_out, num_workers, scheduler, progressbar)


class Pipeline:
    def __init__(self, functions: List[TransformFunction]):
        """Takes a list of functions decorated using hub.compute and creates a pipeline that can be evaluated using .eval"""
        self.functions = functions

    def __len__(self):
        return len(self.functions)

    def eval(
        self,
        data_in,
        ds_out: hub.Dataset,
        num_workers: int = 0,
        scheduler: str = "threaded",
        progressbar: bool = True,
    ):
        """Evaluates the pipeline on data_in to produce an output dataset ds_out.

        Args:
            data_in: Input passed to the transform to generate output dataset. Should support __getitem__ and __len__. Can be a Hub dataset.
            ds_out (Dataset): The dataset object to which the transform will get written.
                Should have all keys being generated in output already present as tensors. It's initial state should be either:-
                - Empty i.e. all tensors have no samples. In this case all samples are added to the dataset.
                - All tensors are populated and have sampe length. In this case new samples are appended to the dataset.
            num_workers (int): The number of workers to use for performing the transform. Defaults to 0. When set to 0, it will always use serial processing, irrespective of the scheduler.
            scheduler (str): The scheduler to be used to compute the transformation. Supported values include: "serial", 'threaded', 'processed' and 'ray'.
            progressbar (bool): Displays a progress bar if True (default).

        Raises:
            InvalidInputDataError: If data_in passed to transform is invalid. It should support __getitem__ and __len__ operations. Using scheduler other than "threaded" with hub dataset having base storage as memory as data_in will also raise this.
            InvalidOutputDatasetError: If all the tensors of ds_out passed to transform don't have the same length. Using scheduler other than "threaded" with hub dataset having base storage as memory as ds_out will also raise this.
            TensorMismatchError: If one or more of the outputs generated during transform contain different tensors than the ones present in 'ds_out' provided to transform.
            UnsupportedSchedulerError: If the scheduler passed is not recognized. Supported values include: "serial", 'threaded', 'processed' and 'ray'.
            TransformError: All other exceptions raised if there are problems while running the pipeline.
        """
        num_workers = max(num_workers, 0)
        if num_workers == 0:
            scheduler = "serial"

        if isinstance(data_in, hub.Dataset):
            data_in = get_dataset_with_zero_size_cache(data_in)

        hub_reporter.feature_report(
            feature_name="eval",
            parameters={"Num_Workers": str(num_workers), "Scheduler": scheduler},
        )

        check_transform_data_in(data_in, scheduler)
        check_transform_ds_out(ds_out, scheduler)

        ds_out.flush()
        initial_autoflush = ds_out.storage.autoflush
        ds_out.storage.autoflush = False

        tensors = list(ds_out.tensors)

        compute_provider = get_compute_provider(scheduler, num_workers)

        try:
            self.run(
                data_in, ds_out, tensors, compute_provider, num_workers, progressbar
            )
        except Exception as e:
            raise TransformError(e)
        finally:
            compute_provider.close()

        ds_out.storage.autoflush = initial_autoflush

    def _run_with_progbar(
        self, func: Callable, ret: dict, total: int, desc: Optional[str] = ""
    ):
        """
        Args:
            func (Callable): Function to be executed
            ret (dict): `func` should place its return value in this dictionary
            total (int): Total number of steps in the progress bar
            desc (str, Optional): Description for the progress bar

        Raises:
            Exception: If any worker encounters an error, it is raised
        """
        ismac = sys.platform == "darwin"
        progress = {"value": 0, "error": None}

        def callback(data):
            if isinstance(data, str):
                progress["error"] = data
            else:
                progress["value"] += data

        server = Server(callback)
        port = server.port
        thread = threading.Thread(target=func, args=(port,), daemon=ismac)
        thread.start()
        try:
            for i in tqdm(range(total), desc=desc):
                while i + 1 > progress["value"]:
                    time.sleep(1)
                    err = progress["error"]
                    if err:
                        raise Exception(err)
        finally:
            if ismac:
                while not ret:  # thread.join() takes forever on mac
                    time.sleep(1)
            else:
                thread.join()
            server.stop()

    def run(
        self,
        data_in,
        ds_out: hub.Dataset,
        tensors: List[str],
        compute: ComputeProvider,
        num_workers: int,
        progressbar: bool = True,
    ):
        """Runs the pipeline on the input data to produce output samples and stores in the dataset.
        This receives arguments processed and sanitized by the Pipeline.eval method.
        """
        is_serial = isinstance(compute, SerialProvider)
        num_workers = max(num_workers, 1)
        size = math.ceil(len(data_in) / num_workers)
        slices = [data_in[i * size : (i + 1) * size] for i in range(num_workers)]

        output_base_storage = get_base_storage(ds_out.storage)
        version_state = ds_out.version_state
        tensors = [ds_out.tensors[t].key for t in tensors]

        ret = {}

        def _run(progress_port=None):
            ret["metas_and_encoders"] = compute.map(
                store_data_slice,
                zip(
                    slices,
                    repeat((output_base_storage, ds_out.group_index)),  # type: ignore
                    repeat(tensors),
                    repeat(self),
                    repeat(version_state),
                    repeat(progress_port),
                ),
            )

        if progressbar:
            self._run_with_progbar(
                _run, ret, len(data_in), get_pbar_description(self.functions)
            )
        else:
            _run()

        metas_and_encoders = ret["metas_and_encoders"]

        all_tensor_metas, all_chunk_id_encoders = zip(*metas_and_encoders)
        merge_all_tensor_metas(all_tensor_metas, ds_out)
        merge_all_chunk_id_encoders(all_chunk_id_encoders, ds_out)


def compose(functions: List[TransformFunction]):
    """Takes a list of functions decorated using hub.compute and creates a pipeline that can be evaluated using .eval"""
    if not functions:
        raise HubComposeEmptyListError
    for index, fn in enumerate(functions):
        if not isinstance(fn, TransformFunction):
            raise HubComposeIncompatibleFunction(index)
    return Pipeline(functions)


def compute(fn):
    """Compute is a decorator for functions.
    The functions should have atleast 2 argument, the first two will correspond to sample_in and samples_out.
    There can be as many other arguments as required.
    The output should be appended/extended to the second argument in a hub like syntax.
    Any value returned by the fn will be ignored.

    Example::

        @hub.compute
        def your_function(sample_in: Any, samples_out, your_arg0, your_arg1=0):
            samples_out.your_tensor.append(your_arg0 * your_arg1)
    """

    def inner(*args, **kwargs):
        return TransformFunction(fn, args, kwargs)

    return inner



from hub.constants import LOCAL_CACHE_PREFIX
from typing import List, Optional
from uuid import uuid1

from hub.core.storage import StorageProvider, MemoryProvider, LocalProvider
from hub.core.storage.lru_cache import LRUCache
from hub.util.exceptions import ProviderSizeListMismatch, ProviderListEmptyError


def get_cache_chain(storage_list: List[StorageProvider], size_list: List[int]):
    """Returns a chain of storage providers as a cache

    Args:
        storage_list (List[StorageProvider]): The list of storage providers needed in a cache.
            Should have atleast one provider in the list.
            If only one provider, LRU cache isn't created and the provider is returned.
        size_list (List[int]): The list of sizes of the caches in bytes.
            Should have size 1 less than provider_list and specifies size of cache for all providers except the last
            one. The last one is the primary storage and is assumed to have infinite space.

    Returns:
        StorageProvider: Returns a cache containing all the storage providers in cache_list if cache_list has 2 or more
            elements.
            Returns the provider if the provider_list has only one provider.

    Raises:
        ProviderListEmptyError: If the provider list is empty.
        ProviderSizeListMismatch: If the len(size_list) + 1 != len(provider_list)
    """
    if not storage_list:
        raise ProviderListEmptyError
    if len(storage_list) <= 1:
        return storage_list[0]
    if len(size_list) + 1 != len(storage_list):
        raise ProviderSizeListMismatch
    store = storage_list[-1]
    for size, cache in zip(reversed(size_list), reversed(storage_list[:-1])):
        store = LRUCache(cache, store, size)
    return store


def generate_chain(
    base_storage: StorageProvider,
    memory_cache_size: int,
    local_cache_size: int,
    path: Optional[str] = None,
) -> StorageProvider:
    """Internal function to be used by Dataset, to generate a cache_chain using a base_storage and sizes of memory and
        local caches.

    Args:
        base_storage (StorageProvider): The underlying actual storage of the Dataset.
        memory_cache_size (int): The size of the memory cache to be used in bytes.
        local_cache_size (int): The size of the local filesystem cache to be used in bytes.
        path (str, optional): The path to the dataset. If not None, it is used to figure out the folder name where the local
            cache is stored.

    Returns:
        StorageProvider: Returns a cache containing the base_storage along with memory cache,
            and local cache if a positive size has been specified for it.
    """

    if path:
        cached_dataset_name = path.replace("://", "_")
        cached_dataset_name = cached_dataset_name.replace("/", "_")
        cached_dataset_name = cached_dataset_name.replace("\\", "_")
    else:
        cached_dataset_name = str(uuid1())

    storage_list: List[StorageProvider] = []
    size_list: List[int] = []

    # Always have a memory cache prefix. Required for support for Cachable objects.
    storage_list.append(MemoryProvider(f"cache/{cached_dataset_name}"))
    size_list.append(memory_cache_size)

    if local_cache_size > 0:
        storage_list.append(
            LocalProvider(f"{LOCAL_CACHE_PREFIX}/{cached_dataset_name}")
        )
        size_list.append(local_cache_size)
    storage_list.append(base_storage)
    return get_cache_chain(storage_list, size_list)



from collections import OrderedDict
from hub.core.storage.cachable import Cachable, CachableCallback
from typing import Any, Dict, Optional, Set, Union

from hub.core.storage.provider import StorageProvider


def _get_nbytes(obj: Union[bytes, memoryview, Cachable]):
    if isinstance(obj, Cachable):
        return obj.nbytes
    return len(obj)


# TODO use lock for multiprocessing
class LRUCache(StorageProvider):
    """LRU Cache that uses StorageProvider for caching"""

    def __init__(
        self,
        cache_storage: StorageProvider,
        next_storage: Optional[StorageProvider],
        cache_size: int,
    ):
        """Initializes the LRUCache. It can be chained with other LRUCache objects to create multilayer caches.

        Args:
            cache_storage (StorageProvider): The storage being used as the caching layer of the cache.
                This should be a base provider such as MemoryProvider, LocalProvider or S3Provider but not another LRUCache.
            next_storage (StorageProvider): The next storage layer of the cache.
                This can either be a base provider (i.e. it is the final storage) or another LRUCache (i.e. in case of chained cache).
                While reading data, all misses from cache would be retrieved from here.
                While writing data, the data will be written to the next_storage when cache_storage is full or flush is called.
            cache_size (int): The total space that can be used from the cache_storage in bytes.
                This number may be less than the actual space available on the cache_storage.
                Setting it to a higher value than actually available space may lead to unexpected behaviors.
        """
        self.next_storage = next_storage
        self.cache_storage = cache_storage
        self.cache_size = cache_size

        # tracks keys in lru order, stores size of value, only keys present in this exist in cache
        self.lru_sizes: OrderedDict[str, int] = OrderedDict()
        self.dirty_keys: Set[str] = set()  # keys present in cache but not next_storage
        self.cache_used = 0

    def update_used_cache_for_path(self, path: str, new_size: int):
        if new_size < 0:
            raise ValueError(f"`new_size` must be >= 0. Got: {new_size}")
        if path in self.lru_sizes:
            old_size = self.lru_sizes[path]
            self.cache_used -= old_size
        self.cache_used += new_size
        self.lru_sizes[path] = new_size

    def flush(self):
        """Writes data from cache_storage to next_storage. Only the dirty keys are written.
        This is a cascading function and leads to data being written to the final storage in case of a chained cache.
        """
        self.check_readonly()
        if self.dirty_keys:
            for key in self.dirty_keys.copy():
                self._forward(key)
            if self.next_storage is not None:
                self.next_storage.flush()

    def get_cachable(self, path: str, expected_class):
        """If the data at `path` was stored using the output of a `Cachable` object's `tobytes` function,
        this function will read it back into object form & keep the object in cache.

        Args:
            path (str): Path to the stored cachable.
            expected_class (callable): The expected subclass of `Cachable`.

        Raises:
            ValueError: If the incorrect `expected_class` was provided.
            ValueError: If the type of the data at `path` is invalid.

        Returns:
            An instance of `expected_class` populated with the data.
        """

        item = self[path]

        if isinstance(item, Cachable):
            if type(item) != expected_class:
                raise ValueError(
                    f"'{path}' was expected to have the class '{expected_class.__name__}'. Instead, got: '{type(item)}'."
                )
            return item

        if isinstance(item, (bytes, memoryview)):
            obj = expected_class.frombuffer(item)

            if isinstance(obj, CachableCallback):
                obj.initialize_callback_location(path, self)

            if obj.nbytes <= self.cache_size:
                self._insert_in_cache(path, obj)

            return obj

        raise ValueError(f"Item at '{path}' got an invalid type: '{type(item)}'.")

    def __getitem__(self, path: str):
        """If item is in cache_storage, retrieves from there and returns.
        If item isn't in cache_storage, retrieves from next storage, stores in cache_storage (if possible) and returns.

        Args:
            path (str): The path relative to the root of the underlying storage.

        Raises:
            KeyError: if an object is not found at the path.

        Returns:
            bytes: The bytes of the object present at the path.
        """
        if path in self.lru_sizes:
            self.lru_sizes.move_to_end(path)  # refresh position for LRU
            return self.cache_storage[path]
        else:
            if self.next_storage is not None:
                # fetch from storage, may throw KeyError
                result = self.next_storage[path]

                if _get_nbytes(result) <= self.cache_size:  # insert in cache if it fits
                    self._insert_in_cache(path, result)
                return result
            raise KeyError(path)

    def __setitem__(self, path: str, value: Union[bytes, Cachable]):
        """Puts the item in the cache_storage (if possible), else writes to next_storage.

        Args:
            path (str): the path relative to the root of the underlying storage.
            value (bytes): the value to be assigned at the path.

        Raises:
            ReadOnlyError: If the provider is in read-only mode.
        """
        self.check_readonly()
        if path in self.lru_sizes:
            size = self.lru_sizes.pop(path)
            self.cache_used -= size

        if _get_nbytes(value) <= self.cache_size:
            self._insert_in_cache(path, value)
            self.dirty_keys.add(path)
        else:  # larger than cache, directly send to next layer
            self._forward_value(path, value)

        self.maybe_flush()

    def __delitem__(self, path: str):
        """Deletes the object present at the path from the cache and the underlying storage.

        Args:
            path (str): the path to the object relative to the root of the provider.

        Raises:
            KeyError: If an object is not found at the path.
            ReadOnlyError: If the provider is in read-only mode.
        """
        self.check_readonly()
        deleted_from_cache = False
        if path in self.lru_sizes:
            size = self.lru_sizes.pop(path)
            self.cache_used -= size
            del self.cache_storage[path]
            self.dirty_keys.discard(path)
            deleted_from_cache = True

        try:
            if self.next_storage is not None:
                del self.next_storage[path]
            else:
                raise KeyError(path)
        except KeyError:
            if not deleted_from_cache:
                raise

        self.maybe_flush()

    def clear_cache(self):
        """Flushes the content of all the cache layers if not in read mode and and then deletes contents of all the layers of it.
        This doesn't delete data from the actual storage.
        """
        self._flush_if_not_read_only()
        self.cache_used = 0
        self.lru_sizes.clear()
        self.dirty_keys.clear()
        self.cache_storage.clear()

        if self.next_storage is not None and hasattr(self.next_storage, "clear_cache"):
            self.next_storage.clear_cache()

    def clear(self):
        """Deletes ALL the data from all the layers of the cache and the actual storage.
        This is an IRREVERSIBLE operation. Data once deleted can not be recovered.
        """
        self.check_readonly()
        self.cache_used = 0
        self.lru_sizes.clear()
        self.dirty_keys.clear()
        self.cache_storage.clear()
        if self.next_storage is not None:
            self.next_storage.clear()

    def __len__(self):
        """Returns the number of files present in the cache and the underlying storage.

        Returns:
            int: the number of files present inside the root.
        """
        return len(self._all_keys())

    def __iter__(self):
        """Generator function that iterates over the keys of the cache and the underlying storage.

        Yields:
            str: the path of the object that it is iterating over, relative to the root of the provider.
        """
        yield from self._all_keys()

    def _forward(self, path, remove_from_dirty=False):
        """Forward the value at a given path to the next storage, and un-marks its key.
        If the value at the path is Cachable, it will only be un-dirtied if remove_from_dirty=True.
        """
        if self.next_storage is not None:
            self._forward_value(path, self.cache_storage[path], remove_from_dirty)

    def _forward_value(self, path, value, remove_from_dirty=False):
        """Forwards a path-value pair to the next storage, and un-marks its key.

        Args:
            path (str): the path to the object relative to the root of the provider.
            value (bytes, Cachable): the value to send to the next storage.
            remove_from_dirty (bool, optional): cachable values are not un-marked automatically,
                as they are externally mutable. Set this to True to un-mark them anyway.
        """
        if self.next_storage is not None:
            cachable = isinstance(value, Cachable)

            if not cachable or remove_from_dirty:
                self.dirty_keys.discard(path)

            if cachable:
                self.next_storage[path] = value.tobytes()
            else:
                self.next_storage[path] = value

    def _free_up_space(self, extra_size: int):
        """Helper function that frees up space the requred space in cache.
            No action is taken if there is sufficient space in the cache.

        Args:
            extra_size (int): the space that needs is required in bytes.
        """
        while self.cache_used > 0 and extra_size + self.cache_used > self.cache_size:
            self._pop_from_cache()

    def _pop_from_cache(self):
        """Helper function that pops the least recently used key, value pair from the cache"""
        key, itemsize = self.lru_sizes.popitem(last=False)
        if key in self.dirty_keys:
            self._forward(key, remove_from_dirty=True)
        del self.cache_storage[key]
        self.cache_used -= itemsize

    def _insert_in_cache(self, path: str, value: Union[bytes, Cachable]):
        """Helper function that adds a key value pair to the cache.

        Args:
            path (str): the path relative to the root of the underlying storage.
            value (bytes): the value to be assigned at the path.

        Raises:
            ReadOnlyError: If the provider is in read-only mode.
        """

        self._free_up_space(_get_nbytes(value))
        self.cache_storage[path] = value  # type: ignore

        self.update_used_cache_for_path(path, _get_nbytes(value))

    def _all_keys(self):
        """Helper function that lists all the objects present in the cache and the underlying storage.

        Returns:
            set: set of all the objects found in the cache and the underlying storage.
        """
        key_set = set()
        if self.next_storage is not None:
            key_set = self.next_storage._all_keys()  # type: ignore
        key_set = key_set.union(self.cache_storage._all_keys())
        return key_set

    def _flush_if_not_read_only(self):
        """Flushes the cache if not in read-only mode."""
        if not self.read_only:
            self.flush()

    def __getstate__(self) -> Dict[str, Any]:
        """Returns the state of the cache, for pickling"""

        # flushes the cache before pickling
        self._flush_if_not_read_only()

        return {
            "next_storage": self.next_storage,
            "cache_storage": self.cache_storage,
            "cache_size": self.cache_size,
        }

    def __setstate__(self, state: Dict[str, Any]):
        """Recreates a cache with the same configuration as the state.

        Args:
            state (dict): The state to be used to recreate the cache.

        Note:
            While restoring the cache, we reset its contents.
            In case the cache storage was local/s3 and is still accessible when unpickled (if same machine/s3 creds present respectively), the earlier cache contents are no longer accessible.
        """

        # TODO: We might want to change this behaviour in the future by having a separate file that keeps a track of the lru order for restoring the cache.
        # This would also allow the cache to persist across different different Dataset objects pointing to the same dataset.
        self.next_storage = state["next_storage"]
        self.cache_storage = state["cache_storage"]
        self.cache_size = state["cache_size"]
        self.lru_sizes = OrderedDict()
        self.dirty_keys = set()
        self.cache_used = 0



import hub
import warnings
import json
import numpy as np
from math import ceil
from typing import Any, Dict, Optional, Sequence, Union, Tuple, List, Set

from hub.compression import get_compression_type, BYTE_COMPRESSION, IMAGE_COMPRESSION
from hub.core.version_control.commit_node import CommitNode  # type: ignore
from hub.core.version_control.commit_chunk_set import CommitChunkSet  # type: ignore
from hub.core.fast_forwarding import ffw_chunk_id_encoder
from hub.core.compression import decompress_array, decompress_bytes
from hub.core.sample import Sample, SampleValue  # type: ignore
from hub.core.meta.tensor_meta import TensorMeta
from hub.core.index.index import Index
from hub.core.storage.lru_cache import LRUCache
from hub.core.chunk import Chunk
from hub.core.meta.encode.chunk_id import ChunkIdEncoder
from hub.core.serialize import serialize_input_samples
from hub.core.compression import compress_multiple, decompress_multiple
from hub.constants import DEFAULT_MAX_CHUNK_SIZE, FIRST_COMMIT_ID


from hub.util.keys import (
    get_chunk_key,
    get_chunk_id_encoder_key,
    get_tensor_meta_key,
    get_tensor_commit_chunk_set_key,
)
from hub.util.exceptions import (
    CorruptedMetaError,
    DynamicTensorNumpyError,
)
from hub.util.json import HubJsonDecoder
from hub.util.casting import get_dtype, intelligent_cast
from hub.util.version_control import auto_checkout, commit, commit_chunk_set_exists


def is_uniform_sequence(samples):
    """Determines if a sequence of samples has uniform type and shape, allowing it to be vectorized by `ChunkEngine.extend`."""
    if len(set(map(type, samples))) != 1:
        # Cannot vectorize sequence with inconsistent types
        return False
    elif any(isinstance(s, np.ndarray) for s in samples):
        # Numpy arrays will only be vectorized if they have the same shape
        return len(set(s.shape for s in samples)) == 1
    elif any(isinstance(s, Sample) for s in samples):
        # Sample objects will not be vectorized
        return False
    else:
        # Scalar samples can be vectorized
        return True


# used for warning the user if updating a tensor caused suboptimal chunks
CHUNK_UPDATE_WARN_PORTION = 0.2


class ChunkEngine:
    def __init__(
        self,
        key: str,
        cache: LRUCache,
        version_state: Dict[str, Any],
        meta_cache: LRUCache = None,
    ):
        """Handles creating `Chunk`s and filling them with incoming samples.

        Data delegation:
            All samples must live inside a chunk. No chunks may contain partial samples, only 1 chunk per sample.
            A chunk holds the dynamic information for the samples they contain (like shape and byte ranges).
            For more information on the `Chunk` format, check out the `Chunk` class.

        ChunkIdEncoder:
            The `ChunkIdEncoder` bidirectionally maps samples to the chunk IDs they live in. For more information,
            see `ChunkIdEncoder`'s docstring.

        Example:
            Given:
                Sample sizes: [1 * MB, 1 * MB, 14 * MB, 15 * MB, 15 * MB]
                Min chunk size: 16 * MB
                Max chunk size: 32 * MB


            Basic logic:
                >>> chunks = []
                >>> chunks.append(sum([1 * MB, 1 * MB, 14 * MB, 15 * MB]))  # i=(0, 1, 2, 3)
                >>> chunks[-1]
                31 * MB
                >>> chunks.append(sum([15 * MB]))  # i=(4,)
                >>> chunks[-1]
                15 * MB

            Samples 0, 1, 2, and 3 can be stored in 1 chunk. sample 4 resides in it's own chunk.

            If more samples come later: sizes = [15 * MB, 1 * MB]

            Basic logic:
                >>> len(chunks)
                2
                >>> chunks[-1]
                15 * MB
                >>> chunks[-1] += sum([15 * MB, 1 * MB])  # i=(5, 6)
                >>> chunks[-1]
                31 * MB
                >>> sum(chunks)
                62 * MB
                >>> len(chunks)
                2

            Because our max chunk size is 32 * MB, we try to fit as much data into this size as possible.


        Args:
            key (str): Tensor key.
            cache (LRUCache): Cache for which chunks and the metadata are stored.
            version_state (Dict[str, Any]): The version state of the dataset, includes commit_id, commit_node, branch, branch_commit_map and commit_node_map.
            meta_cache (LRUCache): Cache used for storing non chunk data such as tensor meta and chunk id encoder during transforms in memory.

        Raises:
            ValueError: If invalid max chunk size.
        """

        self.key = key
        self.cache = cache
        self._meta_cache = meta_cache
        self.version_state = version_state
        self._last_row = 0

        if self.tensor_meta.chunk_compression:
            # Cache samples in the last chunk in uncompressed form.
            self._last_chunk_uncompressed: List[np.ndarray] = (
                self.last_chunk.decompressed_samples(
                    compression=self.tensor_meta.chunk_compression,
                    dtype=self.tensor_meta.dtype,
                )
                if self.last_chunk
                else []
            )

    @property
    def max_chunk_size(self):
        # no chunks may exceed this
        return (
            getattr(self.tensor_meta, "max_chunk_size", None) or DEFAULT_MAX_CHUNK_SIZE
        )

    @property
    def min_chunk_size(self):
        # only the last chunk may be less than this
        return self.max_chunk_size // 2

    @property
    def meta_cache(self) -> LRUCache:
        return self._meta_cache or self.cache

    @property
    def chunk_id_encoder(self) -> ChunkIdEncoder:
        """Gets the chunk id encoder from cache, if one is not found it creates a blank encoder.
        For more information on what `ChunkIdEncoder` is used for, see the `__init__` docstring.

        Raises:
            CorruptedMetaError: If chunk id encoding was corrupted.

        Returns:
            ChunkIdEncoder: The chunk ID encoder handles the mapping between sample indices
                and their corresponding chunks.
        """
        commit_id = self.version_state["commit_id"]
        key = get_chunk_id_encoder_key(self.key, commit_id)
        if not self.chunk_id_encoder_exists:
            enc = ChunkIdEncoder()
            self.meta_cache[key] = enc
            return enc

        enc = self.meta_cache.get_cachable(key, ChunkIdEncoder)
        return enc

    @property
    def commit_chunk_set(self) -> Optional[CommitChunkSet]:
        """Gets the commit chunk set from cache, if one is not found it creates a blank one.

        Returns:
            Optional[CommitChunkSet]: The commit chunk set keeps track of all the chunks present in the current commit, returns None for the first commit.
        """
        commit_id = self.version_state["commit_id"]
        if commit_id == FIRST_COMMIT_ID:
            # the first commit doesn't need a commit chunk set
            return None
        key = get_tensor_commit_chunk_set_key(self.key, commit_id)
        if not self.commit_chunk_set_exists:
            cset = CommitChunkSet()
            self.meta_cache[key] = cset
            return cset

        cset = self.meta_cache.get_cachable(key, CommitChunkSet)
        return cset

    @property
    def commit_chunk_set_exists(self) -> bool:
        return commit_chunk_set_exists(self.version_state, self.meta_cache, self.key)

    @property
    def chunk_id_encoder_exists(self) -> bool:
        try:
            commit_id = self.version_state["commit_id"]
            key = get_chunk_id_encoder_key(self.key, commit_id)
            self.meta_cache[key]
            return True
        except KeyError:
            return False

    @property
    def num_chunks(self) -> int:
        if not self.chunk_id_encoder_exists:
            return 0
        return self.chunk_id_encoder.num_chunks

    @property
    def num_samples(self) -> int:
        if not self.chunk_id_encoder_exists:
            return 0
        return self.chunk_id_encoder.num_samples

    @property
    def last_chunk(self) -> Optional[Chunk]:
        if self.num_chunks == 0:
            return None
        chunk_name = self.last_chunk_name
        chunk_commit_id = self.get_chunk_commit(chunk_name)
        chunk_key = get_chunk_key(self.key, chunk_name, chunk_commit_id)
        chunk = self.get_chunk(chunk_key)
        if chunk_commit_id != self.version_state["commit_id"]:
            chunk = self.copy_chunk_to_new_commit(chunk, chunk_name)
        chunk.key = chunk_key  # type: ignore
        return chunk

    def get_chunk(self, chunk_key: str) -> Chunk:
        chunk = self.cache.get_cachable(chunk_key, Chunk)
        chunk.key = chunk_key
        return chunk

    def get_chunk_commit(self, chunk_name) -> str:
        """Returns the commit id that contains the chunk_name."""
        cur_node: Optional[CommitNode] = self.version_state["commit_node"]
        while cur_node is not None:
            commit_id = cur_node.commit_id
            chunk_set_key = get_tensor_commit_chunk_set_key(self.key, commit_id)
            try:
                # the first commit doesn't contain a chunk set, don't repeatedly try to fetch from storage
                if commit_id == FIRST_COMMIT_ID:
                    chunk_set = set()
                else:
                    chunk_set = self.meta_cache.get_cachable(
                        chunk_set_key, CommitChunkSet
                    ).chunks
            except Exception:
                chunk_set = set()
            if chunk_name in chunk_set:
                return commit_id
            cur_node = cur_node.parent  # type: ignore
        # the first commit doesn't have a commit chunk set, so any chunk that wasn't found belongs to the first commit
        return FIRST_COMMIT_ID

    @property
    def last_chunk_key(self) -> str:
        last_chunk_name = self.last_chunk_name
        commit_id = self.get_chunk_commit(last_chunk_name)
        return get_chunk_key(self.key, last_chunk_name, commit_id)

    @property
    def last_chunk_name(self) -> str:
        return self.chunk_id_encoder.get_name_for_chunk(-1)

    @property
    def tensor_meta(self):
        tensor_meta_key = get_tensor_meta_key(self.key, self.version_state["commit_id"])
        return self.meta_cache.get_cachable(tensor_meta_key, TensorMeta)

    def _extend_bytes(
        self,
        buffer: memoryview,
        nbytes: List[int],
        shapes: List[Tuple[int]],
    ) -> List[Chunk]:
        """Treat `buffer` as multiple samples and place them into compressed `Chunk`s."""
        if self.tensor_meta.chunk_compression:
            raise NotImplementedError(
                "_extend_bytes not implemented for tensors with chunk wise compression. Use _append_bytes instead."
            )
        chunk = self.last_chunk
        new_chunk = self._create_new_chunk
        if chunk is None:
            chunk = new_chunk()

        updated_chunks = set()

        # If the first incoming sample can't fit in the last chunk, create a new chunk.
        if nbytes[0] > self.min_chunk_size - chunk.num_data_bytes:
            chunk = self._create_new_chunk()

        max_chunk_size = self.max_chunk_size
        min_chunk_size = self.min_chunk_size
        enc = self.chunk_id_encoder

        while nbytes:  # len(nbytes) is initially same as number of incoming samples.
            num_samples_to_current_chunk = 0
            nbytes_to_current_chunk = 0
            for nb in nbytes:  # len(nbytes) = samples remaining to be added to a chunk

                # Size of the current chunk if this sample is added to it
                chunk_future_size = nbytes_to_current_chunk + nb + chunk.num_data_bytes  # type: ignore
                if chunk_future_size > max_chunk_size:
                    break

                num_samples_to_current_chunk += 1
                nbytes_to_current_chunk += nb
                if (
                    chunk_future_size > min_chunk_size
                ):  # Try to keep chunk size close to min_chunk_size
                    break
            chunk.extend_samples(  # type: ignore
                buffer[:nbytes_to_current_chunk],
                max_chunk_size,
                shapes[:num_samples_to_current_chunk],
                nbytes[:num_samples_to_current_chunk],
            )
            updated_chunks.add(chunk)
            enc.register_samples(num_samples_to_current_chunk)

            # Remove bytes from buffer that have been added to current chunk
            buffer = buffer[nbytes_to_current_chunk:]

            # Remove shapes and nbytes for samples that have beed added to current chunk
            del nbytes[:num_samples_to_current_chunk]
            del shapes[:num_samples_to_current_chunk]

            if buffer:
                chunk = new_chunk()
        return updated_chunks  # type: ignore

    def _append_bytes_to_compressed_chunk(self, buffer: memoryview, shape: Tuple[int]):
        """Treat `buffer` as single sample and place them into compressed `Chunk`s."""
        chunk_compression = self.tensor_meta.chunk_compression
        last_chunk_uncompressed = self._last_chunk_uncompressed

        # Append incoming buffer to last chunk and compress:
        last_chunk_uncompressed.append(
            np.frombuffer(buffer, dtype=self.tensor_meta.dtype).reshape(shape)
        )
        compressed_bytes = compress_multiple(last_chunk_uncompressed, chunk_compression)

        # Check if last chunk can hold new compressed buffer.
        if self._can_set_to_last_chunk(len(compressed_bytes)):
            chunk = self.last_chunk
        else:
            # Last chunk full, create new chunk
            chunk = self._create_new_chunk()

            # All samples except the last one are already in the previous chunk, so remove them from cache and compress:
            del last_chunk_uncompressed[:-1]
            compressed_bytes = compress_multiple(
                last_chunk_uncompressed, chunk_compression
            )

        # Set chunk data
        chunk._data = compressed_bytes  # type: ignore

        # Update headers
        if get_compression_type(chunk_compression) == BYTE_COMPRESSION:
            chunk.register_sample_to_headers(incoming_num_bytes=len(buffer), sample_shape=shape)  # type: ignore
        else:
            # Byte positions are not relevant for image compressions, so incoming_num_bytes=None.
            chunk.register_sample_to_headers(incoming_num_bytes=None, sample_shape=shape)  # type: ignore
        return chunk

    def _append_bytes(self, buffer: memoryview, shape: Tuple[int]) -> Chunk:
        """Treat `buffer` as a single sample and place them into `Chunk`s. This function implements the algorithm for
        determining which chunks contain which parts of `buffer`.

        Args:
            buffer (memoryview): Buffer that represents a single sample. Can have a
                length of 0, in which case `shape` should contain at least one 0 (empty sample).
            shape (Tuple[int]): Shape for the sample that `buffer` represents.

        Returns:
            Chunk to which the sample was appended.
        """

        # num samples is always 1 when appending
        num_samples = 1

        if self.tensor_meta.chunk_compression:
            chunk_after_append = self._append_bytes_to_compressed_chunk(buffer, shape)
        else:
            chunk_after_append = self._try_appending_to_last_chunk(buffer, shape)
            if not chunk_after_append:
                chunk_after_append = self._append_to_new_chunk(buffer, shape)

        self.chunk_id_encoder.register_samples(num_samples)
        return chunk_after_append

    def _can_set_to_last_chunk(self, nbytes: int) -> bool:
        """Whether last chunk's data can be set to a buffer of size nbytes."""
        last_chunk = self.last_chunk
        if last_chunk is None:
            return False
        return nbytes <= self.min_chunk_size

    def _synchronize_cache(self, chunk_keys: List[str] = None):
        """Synchronizes cachables with the cache.

        Args:
            chunk_keys (List[str]): List of chunk keys to be synchronized. If None, only the last chunk will be synchronized. Defaults to None.
        """

        # TODO implement tests for cache size compute
        # TODO: optimize this by storing all of these keys in the chunk engine's state (posixpath.joins are pretty slow)

        # synchronize chunks
        if chunk_keys is None:
            chunk_keys = [self.last_chunk_key]
        for chunk_key in chunk_keys:
            chunk = self.get_chunk(chunk_key)
            self.cache.update_used_cache_for_path(chunk_key, chunk.nbytes)  # type: ignore

        commit_id = self.version_state["commit_id"]

        # synchronize tensor meta
        tensor_meta_key = get_tensor_meta_key(self.key, commit_id)
        self.meta_cache[tensor_meta_key] = self.tensor_meta

        # synchronize chunk ID encoder
        chunk_id_key = get_chunk_id_encoder_key(self.key, commit_id)
        self.meta_cache[chunk_id_key] = self.chunk_id_encoder

        # first commit doesn't have commit chunk set
        if commit_id != FIRST_COMMIT_ID:
            # synchronize current chunk set, all older ones are immutable
            commit_chunk_set_key = get_tensor_commit_chunk_set_key(self.key, commit_id)
            self.meta_cache[commit_chunk_set_key] = self.commit_chunk_set  # type: ignore

    def _try_appending_to_last_chunk(self, buffer: memoryview, shape: Tuple[int]):
        """Will store `buffer` inside of the last chunk if it can.
        It can be stored in the last chunk if it exists and has space for `buffer`.

        Args:
            buffer (memoryview): Data to store. This can represent any number of samples.
            shape (Tuple[int]): Shape for the sample that `buffer` represents.

        Returns:
            Last chunk if `buffer` was successfully written to the last chunk, otherwise False.
        """

        last_chunk = self.last_chunk
        if last_chunk is None:
            return False

        incoming_num_bytes = len(buffer)

        if last_chunk.is_under_min_space(self.min_chunk_size):
            last_chunk_size = last_chunk.num_data_bytes
            chunk_ct_content = _min_chunk_ct_for_data_size(
                self.max_chunk_size, incoming_num_bytes
            )

            extra_bytes = min(incoming_num_bytes, self.max_chunk_size - last_chunk_size)
            combined_chunk_ct = _min_chunk_ct_for_data_size(
                self.max_chunk_size, incoming_num_bytes + last_chunk_size
            )

            # combine if count is same
            if combined_chunk_ct == chunk_ct_content:
                last_chunk.append_sample(
                    buffer[:extra_bytes], self.max_chunk_size, shape
                )
                return last_chunk

        return False

    def _append_to_new_chunk(self, buffer: memoryview, shape: Tuple[int]):
        """Will create a new chunk and store `buffer` inside of it. Assumes that `buffer`'s length is < max chunk size.
        This should be called if `buffer` could not be added to the last chunk.

        Args:
            buffer (memoryview): Data to store. This can represent any number of samples.
            shape (Tuple[int]): Shape for the sample that `buffer` represents.

        Returns:
            New chunk that was created
        """

        # check if `last_chunk_extended` to handle empty samples
        new_chunk = self._create_new_chunk()
        new_chunk.append_sample(buffer, self.max_chunk_size, shape)
        return new_chunk

    def _create_new_chunk(self):
        """Creates and returns a new `Chunk`. Automatically creates an ID for it and puts a reference in the cache."""

        chunk_id = self.chunk_id_encoder.generate_chunk_id()
        chunk = Chunk()
        chunk_name = ChunkIdEncoder.name_from_id(chunk_id)
        chunk_key = get_chunk_key(self.key, chunk_name, self.version_state["commit_id"])
        if self.commit_chunk_set is not None:
            self.commit_chunk_set.add(chunk_name)
        self.cache[chunk_key] = chunk
        chunk.key = chunk_key
        return chunk

    def extend(self, samples: Union[np.ndarray, Sequence[SampleValue]]):
        """Formats a batch of `samples` and feeds them into `_append_bytes`."""

        self.cache.check_readonly()
        # if not the head node, checkout to an auto branch that is newly created
        auto_checkout(self.version_state, self.cache)
        ffw_chunk_id_encoder(self.chunk_id_encoder)

        tensor_meta = self.tensor_meta
        if tensor_meta.dtype is None:
            tensor_meta.set_dtype(get_dtype(samples))

        buff, nbytes, shapes = serialize_input_samples(
            samples, tensor_meta, self.min_chunk_size
        )
        for shape in shapes:
            tensor_meta.update_shape_interval(shape)
        tensor_meta.length += len(samples)
        if tensor_meta.chunk_compression:
            updated_chunks = set()
            for nb, shape in zip(nbytes, shapes):
                chunk = self._append_bytes(buff[:nb], shape[:])  # type: ignore
                updated_chunks.add(chunk)
                buff = buff[nb:]
                updated_chunks.add(chunk)
        else:
            updated_chunks = self._extend_bytes(buff, nbytes, shapes[:])  # type: ignore
        for chunk in updated_chunks:
            self.cache[chunk.key] = chunk  # type: ignore
        self._synchronize_cache(chunk_keys=[])
        self.cache.maybe_flush()

    def append(self, sample: SampleValue):
        """Formats a single `sample` (compresseses/decompresses if applicable) and feeds it into `_append_bytes`."""
        self.extend([sample])

    def update(
        self,
        index: Index,
        samples: Union[Sequence[SampleValue], SampleValue],
        operator: Optional[str] = None,
    ):
        """Update data at `index` with `samples`."""

        if operator is not None:
            return self._update_with_operator(index, samples, operator)

        self.cache.check_readonly()
        # if not the head node, checkout to an auto branch that is newly created
        auto_checkout(self.version_state, self.cache)
        ffw_chunk_id_encoder(self.chunk_id_encoder)

        tensor_meta = self.tensor_meta
        enc = self.chunk_id_encoder
        updated_chunks = set()

        index_length = index.length(self.num_samples)
        samples = _make_sequence(samples, index_length)
        serialized_input_samples = serialize_input_samples(
            samples, tensor_meta, self.min_chunk_size
        )

        chunks_nbytes_after_updates = []
        global_sample_indices = tuple(index.values[0].indices(self.num_samples))
        buffer, nbytes, shapes = serialized_input_samples
        for i, (nb, shape) in enumerate(zip(nbytes, shapes)):
            global_sample_index = global_sample_indices[i]  # TODO!
            chunk = self.get_chunk_for_sample(global_sample_index, enc, copy=True)
            local_sample_index = enc.translate_index_relative_to_chunks(
                global_sample_index
            )
            tensor_meta.update_shape_interval(shape)
            chunk.update_sample(
                local_sample_index,
                buffer[:nb],  # type: ignore
                shape,
                chunk_compression=self.tensor_meta.chunk_compression,
                dtype=self.tensor_meta.dtype,
            )
            buffer = buffer[nb:]
            updated_chunks.add(chunk)

            # only care about deltas if it isn't the last chunk
            if chunk.key != self.last_chunk_key:  # type: ignore
                chunks_nbytes_after_updates.append(chunk.nbytes)

        # TODO: [refactor] this is a hacky way, also `self._synchronize_cache` might be redundant. maybe chunks should use callbacks.
        for chunk in updated_chunks:
            self.cache[chunk.key] = chunk  # type: ignore

        self._synchronize_cache(chunk_keys=[])
        self.cache.maybe_flush()

        _warn_if_suboptimal_chunks(
            chunks_nbytes_after_updates, self.min_chunk_size, self.max_chunk_size
        )

    def _update_with_operator(
        self,
        index: Index,
        samples: Union[Sequence[SampleValue], SampleValue],
        operator: str,
    ):
        """Update data at `index` with the output of elem-wise operatorion with samples"""
        try:
            if isinstance(samples, hub.core.tensor.Tensor):
                samples = samples.numpy()
            arr = self.numpy(index)
        except DynamicTensorNumpyError:
            raise NotImplementedError(
                "Inplace update operations are not available for dynamic tensors yet."
            )
        samples = intelligent_cast(
            samples, self.tensor_meta.dtype, self.tensor_meta.htype
        )
        getattr(arr, operator)(samples)
        self.update(index, arr)

    def numpy(
        self, index: Index, aslist: bool = False
    ) -> Union[np.ndarray, List[np.ndarray]]:
        """Reads samples from chunks and returns as a numpy array. If `aslist=True`, returns a sequence of numpy arrays.

        Args:
            index (Index): Represents the samples to read from chunks. See `Index` for more information.
            aslist (bool): If True, the samples will be returned as a list of numpy arrays. If False, returns a single numpy array. Defaults to False.

        Raises:
            DynamicTensorNumpyError: If shapes of the samples being read are not all the same.

        Returns:
            Union[np.ndarray, Sequence[np.ndarray]]: Either a list of numpy arrays or a single numpy array (depending on the `aslist` argument).
        """
        length = self.num_samples
        enc = self.chunk_id_encoder
        last_shape = None
        samples = []

        for global_sample_index in index.values[0].indices(length):
            chunk = self.get_chunk_for_sample(global_sample_index, enc)
            sample = self.read_sample_from_chunk(global_sample_index, chunk)
            shape = sample.shape

            if not aslist and last_shape is not None:
                if shape != last_shape:
                    raise DynamicTensorNumpyError(self.key, index, "shape")

            samples.append(sample)
            last_shape = shape

        return _format_read_samples(samples, index, aslist)

    def _is_index_in_last_row(self, arr, index) -> bool:
        """Checks if `index` is in the self._last_row of of chunk_id_encoder."""
        row = self._last_row
        return arr[row][1] >= index and (row == 0 or arr[row - 1][1] < index)

    def _get_chunk_id_for_index(self, global_sample_index: int, enc: ChunkIdEncoder):
        """Takes a look at self._last_row and tries to find chunk id without binary search by looking at the current and next row.
        Resorts to binary search if the current and next row don't have global_sample_index in them.
        """
        found = False
        arr = enc.array
        if self._is_index_in_last_row(arr, global_sample_index):
            chunk_id = arr[self._last_row][0]
            found = True
        elif self._last_row < len(arr) - 1:
            self._last_row += 1
            if self._is_index_in_last_row(arr, global_sample_index):
                chunk_id = arr[self._last_row][0]
                found = True

        if not found:
            chunk_id, self._last_row = enc.__getitem__(global_sample_index, True)
        return chunk_id

    def get_chunk_for_sample(
        self, global_sample_index: int, enc: ChunkIdEncoder, copy: bool = False
    ) -> Chunk:
        """Retrives the `Chunk` object corresponding to `global_sample_index`.
        Args:
            global_sample_index (int): Index relative to the entire tensor representing the sample.
            enc (ChunkIdEncoder): Chunk ID encoder. This is an argument because right now it is
                sub-optimal to use `self.chunk_id_encoder` due to posixpath joins.
            copy (bool): If True and the chunk exists in a different commit to the current commit, it will be copied. Defaults to False.
        Returns:
            Chunk: Chunk object that contains `global_sample_index`.
        """
        chunk_id = self._get_chunk_id_for_index(global_sample_index, enc)

        chunk_name = ChunkIdEncoder.name_from_id(chunk_id)
        chunk_commit_id = self.get_chunk_commit(chunk_name)
        current_commit_id = self.version_state["commit_id"]
        chunk_key = get_chunk_key(self.key, chunk_name, chunk_commit_id)
        chunk = self.cache.get_cachable(chunk_key, Chunk)
        chunk.key = chunk_key
        if chunk_commit_id != current_commit_id and copy:
            chunk = self.copy_chunk_to_new_commit(chunk, chunk_name)
        return chunk

    def read_bytes_for_sample(self, global_sample_index: int) -> bytes:
        if self.tensor_meta.chunk_compression:
            raise Exception(
                "Cannot retreive original bytes for samples in chunk-wise compressed tensors."
            )
        enc = self.chunk_id_encoder
        chunk = self.get_chunk_for_sample(global_sample_index, enc)
        buffer = chunk.memoryview_data
        if not buffer:
            return b""
        local_sample_index = enc.translate_index_relative_to_chunks(global_sample_index)
        sb, eb = chunk.byte_positions_encoder[local_sample_index]
        return buffer[sb:eb].tobytes()

    def read_sample_from_chunk(
        self,
        global_sample_index: int,
        chunk: Chunk,
        cast: bool = True,
        copy: bool = False,
    ) -> np.ndarray:
        """Read a sample from a chunk, converts the global index into a local index. Handles decompressing if applicable."""

        dtype = self.tensor_meta.dtype

        enc = self.chunk_id_encoder

        buffer = chunk.memoryview_data

        local_sample_index = enc.translate_index_relative_to_chunks(global_sample_index)
        shape = chunk.shapes_encoder[local_sample_index]

        if len(buffer) == 0:
            return np.zeros(shape, dtype=dtype)

        chunk_compression = self.tensor_meta.chunk_compression
        sample_compression = self.tensor_meta.sample_compression

        htype = self.tensor_meta.htype

        if htype in ("json", "text", "list"):
            sb, eb = chunk.byte_positions_encoder[local_sample_index]
            if chunk_compression:
                decompressed = chunk.decompressed_data(compression=chunk_compression)
                buffer = decompressed[sb:eb]
            elif sample_compression:
                buffer = decompress_bytes(buffer[sb:eb], compression=sample_compression)
            else:
                buffer = buffer[sb:eb]
            buffer = bytes(buffer)
            if htype == "json":
                arr = np.empty(1, dtype=object)
                arr[0] = json.loads(bytes.decode(buffer), cls=HubJsonDecoder)
                return arr
            elif htype == "list":
                lst = json.loads(bytes.decode(buffer), cls=HubJsonDecoder)
                arr = np.empty(len(lst), dtype=object)
                arr[:] = lst
                return arr
            elif htype == "text":
                arr = np.array(bytes.decode(buffer)).reshape(
                    1,
                )
            return arr

        if chunk_compression:
            if get_compression_type(chunk_compression) == BYTE_COMPRESSION:
                decompressed = chunk.decompressed_data(compression=chunk_compression)
                sb, eb = chunk.byte_positions_encoder[local_sample_index]
                return np.frombuffer(decompressed[sb:eb], dtype=dtype).reshape(shape)
            else:
                return chunk.decompressed_samples()[local_sample_index]

        sb, eb = chunk.byte_positions_encoder[local_sample_index]
        buffer = buffer[sb:eb]
        if sample_compression:
            sample = decompress_array(
                buffer, shape, dtype=dtype, compression=sample_compression
            )
            if cast and sample.dtype != dtype:
                sample = sample.astype(dtype)
        else:
            if copy:
                buffer = bytes(buffer)
            sample = np.frombuffer(buffer, dtype=dtype).reshape(shape)

        return sample

    def get_chunk_names_for_multiple_indexes(
        self, sample_index: int, last_index: int, target_chunk_count: int
    ) -> Set[str]:
        """Fetches a set of chunk names in which data starting from sample_index is contained.
            This is used by Pytorch integration.

        Args:
            sample_index: The index starting from which chunk names need to be fetched.
            last_index: The last index till which chunk names need to be fetched.
            target_chunk_count: The target number of chunk names required. The actual size of the returned set may be:-
                a) Less than target_chunk_count: If there are no more chunks to fetch.
                b) More than target_chunk_count: If the last chunk filling up target_chunk_count is a partial chunk, the remaining chunks are fetched.
                c) Equal to the target_chunk_count: In all other cases.

        Returns:
            Set of chunk names.
        """
        chunk_names: Set[str] = set()
        while len(chunk_names) < target_chunk_count and sample_index < last_index:
            chunk_id = self.chunk_id_encoder[sample_index]
            chunk = self.chunk_id_encoder.name_from_id(chunk_id)
            # todo, change to chunk_names.update once chunks returns sequence instead of single string
            chunk_names.add(chunk)
            sample_index += 1
        return chunk_names

    def get_chunk_names_for_index(self, sample_index):
        # TODO: fix this once we support multiple chunk names per sample
        chunk_id = self.chunk_id_encoder[sample_index]
        chunk = self.chunk_id_encoder.name_from_id(chunk_id)
        return [chunk]

    def validate_num_samples_is_synchronized(self):
        """Check if tensor meta length and chunk ID encoder are representing the same number of samples.
        Helpful for determining if a user has tampered with the tensor meta or the chunk ID encoder, or if
        the tensor was corruptd.

        Raises:
            CorruptedMetaError: tensor_meta and chunk_id_encoder must have the same num samples.
        """

        tensor_meta_length = self.tensor_meta.length

        # compare chunk ID encoder and tensor meta
        chunk_id_num_samples = (
            self.chunk_id_encoder.num_samples if self.chunk_id_encoder_exists else 0
        )
        if tensor_meta_length != chunk_id_num_samples:
            commit_id = self.version_state["commit_id"]
            tkey = get_tensor_meta_key(self.key, commit_id)
            ikey = get_chunk_id_encoder_key(self.key, commit_id)
            raise CorruptedMetaError(
                f"'{tkey}' and '{ikey}' have a record of different numbers of samples. Got {tensor_meta_length} and {chunk_id_num_samples} respectively."
            )

    def copy_chunk_to_new_commit(self, chunk, chunk_name):
        """Copies the chunk to the current commit.

        Returns the copied chunk.
        """
        new_chunk_key = get_chunk_key(
            self.key, chunk_name, self.version_state["commit_id"]
        )
        chunk = chunk.copy()
        chunk.key = new_chunk_key
        self.cache[new_chunk_key] = chunk
        if self.commit_chunk_set is not None:
            self.commit_chunk_set.add(chunk_name)
        return chunk


def _format_read_samples(
    samples: Sequence[np.ndarray], index: Index, aslist: bool
) -> Union[np.ndarray, List[np.ndarray]]:
    """Prepares samples being read from the chunk engine in the format the user expects."""

    samples = index.apply(samples)  # type: ignore

    if aslist and all(map(np.isscalar, samples)):
        samples = list(arr.item() for arr in samples)

    samples = index.apply_squeeze(samples)  # type: ignore

    if aslist:
        return samples  # type: ignore
    else:
        return np.array(samples)


def _min_chunk_ct_for_data_size(chunk_max_data_bytes: int, size: int) -> int:
    """Calculates the minimum number of chunks in which data of given size can be fit."""
    return ceil(size / chunk_max_data_bytes)


def _make_sequence(
    samples: Union[Sequence[SampleValue], SampleValue], index_length: int
) -> Sequence[SampleValue]:
    """Make `samples` a sequence of `SampleValue`s.

    Args:
        samples (Union[Sequence[SampleValue], SampleValue]): Incoming samples to be made into a sequence.
        index_length (int): Number of expected samples in the sequence.

    Raises:
        ValueError: If `index_length` is incompatible with the true length of `samples`.

    Returns:
        Sequence[SampleValue]: Sequence of `SampleValue`s with the same length as `index_length`.
    """

    if index_length == 1:
        if hasattr(samples, "__len__"):
            if len(samples) != 1:
                samples = [samples]
        elif hasattr(samples, "shape"):
            if len(samples.shape) > 0 and samples.shape[0] != 1:  # type: ignore
                samples = [samples]
        else:
            samples = [samples]

    if hasattr(samples, "__len__"):
        if index_length != len(samples):
            raise ValueError(
                f"Index length ({index_length}) and length of samples ({len(samples)}) must be equal for updating a tensor."
            )
    else:
        samples = [samples]

    return samples


def _warn_if_suboptimal_chunks(
    chunks_nbytes_after_updates: List[int], min_chunk_size: int, max_chunk_size: int
):
    upper_warn_threshold = max_chunk_size * (1 + CHUNK_UPDATE_WARN_PORTION)
    lower_warn_threshold = min_chunk_size * (1 - CHUNK_UPDATE_WARN_PORTION)

    for nbytes in chunks_nbytes_after_updates:
        if nbytes > upper_warn_threshold or nbytes < lower_warn_threshold:
            warnings.warn(
                "After update, some chunks were suboptimal. Be careful when doing lots of updates that modify the sizes of samples by a large amount, these can heavily impact read performance!"
            )
            break
